from pymatgen.io.vasp.inputs import Poscar
from pymatgen.core.structure import Structure
from matplotlib.colors import LinearSegmentedColormap
from math import floor
from monty.serialization import loadfn, MontyDecoder
import numpy as np
import matplotlib.pyplot as plt
from pymatgen.core.lattice import Lattice
from pymatgen.core.structure import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.analysis.substrate_analyzer import ZSLGenerator, SubstrateAnalyzer
from pymatgen.io.ase import AseAtomsAdaptor
from ase.lattice.surface import surface
from pymatgen.core.surface import Slab, SlabGenerator
from jarvis.db.static.explore_db import get_3d_dataset, get_2d_dataset

plt.switch_backend("agg")


def center_struct(s=""):
    coords = s.cart_coords
    avg_x = np.sum(coords[:, 0]) / len(s)
    avg_y = np.sum(coords[:, 1]) / len(s)
    avg_z = np.sum(coords[:, 2]) / len(s)
    coords = s.cart_coords - [avg_x, avg_y, avg_z]
    new_s = Structure(s.lattice, s.species, coords, coords_are_cartesian=True)
    return new_s


def rangexyz(s=""):
    coords = s.cart_coords
    range_x = abs(max(coords[:, 0]) - min(coords[:, 0]))
    range_y = abs(max(coords[:, 1]) - min(coords[:, 1]))
    range_z = abs(max(coords[:, 2]) - min(coords[:, 2]))
    return range_z


def get_ase_surf(s=[], miller=(0, 0, 1), layers=1):
    strt = center_struct(s)
    ase_atoms = AseAtomsAdaptor().get_atoms(strt)
    ase_slab = surface(ase_atoms, miller, layers)
    ase_slab.center(axis=2)
    tmp = AseAtomsAdaptor().get_structure(ase_slab)
    range_z = rangexyz(strt)
    # print ('rz',abs(strt.lattice.matrix[2][2]),range_z)
    if abs(strt.lattice.matrix[2][2] - range_z) <= 6.0:
        ase_slab.center(axis=2, vacuum=12.0)
    else:
        ase_slab.center(axis=2)
    slab_pymatgen = AseAtomsAdaptor().get_structure(ase_slab)
    slab_pymatgen.sort()
    print("Surface")
    print(Poscar(slab_pymatgen))
    return slab_pymatgen


def mismatch_strts(film=[], subs=[]):

    z = ZSLGenerator()
    matches = list(z(film.lattice.matrix[:2], subs.lattice.matrix[:2], lowest=True))
    info = {}
    info["mismatch_u"] = "na"
    info["mismatch_v"] = "na"
    info["mismatch_angle"] = "na"
    info["area1"] = "na"
    info["area2"] = "na"
    info["film_sl"] = film
    info["subs_sl"] = subs

    try:
        uv1 = matches[0]["sub_sl_vecs"]
        uv2 = matches[0]["film_sl_vecs"]
        # uv1=[[-8.52917200e+00,  9.12745800e+00,  3.66344517e-17],[1.27937580e+01, 9.12745800e+00, 1.34228735e-15]]
        # uv2=[[7.02403800e+00, 1.05360570e+01, 1.07524571e-15],[-1.40480760e+01,  7.02403800e+00, -4.30098283e-16]]
        u = np.array(uv1)
        v = np.array(uv2)
        a1 = u[0]
        a2 = u[1]
        b1 = v[0]
        b2 = v[1]
        mismatch_u = np.linalg.norm(b1) / np.linalg.norm(a1) - 1
        mismatch_v = np.linalg.norm(b2) / np.linalg.norm(a2) - 1
        angle1 = (
            np.arccos(np.dot(a1, a2) / np.linalg.norm(a1) / np.linalg.norm(a2))
            * 180
            / np.pi
        )
        angle2 = (
            np.arccos(np.dot(b1, b2) / np.linalg.norm(b1) / np.linalg.norm(b2))
            * 180
            / np.pi
        )
        mismatch_angle = abs(angle1 - angle2)
        area1 = np.linalg.norm(np.cross(a1, a2))
        area2 = np.linalg.norm(np.cross(b1, b2))
        uv_substrate = uv1
        uv_mat2d = uv2
        substrate_latt = Lattice(
            np.array(
                [uv_substrate[0][:], uv_substrate[1][:], subs.lattice.matrix[2, :]]
            )
        )
        # to avoid numerical issues with find_mapping
        mat2d_fake_c = (
            film.lattice.matrix[2, :] / np.linalg.norm(film.lattice.matrix[2, :]) * 5.0
        )
        mat2d_latt = Lattice(np.array([uv_mat2d[0][:], uv_mat2d[1][:], mat2d_fake_c]))
        mat2d_latt_fake = Lattice(
            np.array(
                [film.lattice.matrix[0, :], film.lattice.matrix[1, :], mat2d_fake_c]
            )
        )
        _, __, scell = subs.lattice.find_mapping(substrate_latt, ltol=0.05, atol=1)
        scell[2] = np.array([0, 0, 1])
        subs.make_supercell(scell)
        _, __, scell = mat2d_latt_fake.find_mapping(mat2d_latt, ltol=0.05, atol=1)
        scell[2] = np.array([0, 0, 1])
        film.make_supercell(scell)
        # modify the substrate lattice so that the 2d material can be
        # grafted on top of it
        lmap = Lattice(
            np.array(
                [
                    subs.lattice.matrix[0, :],
                    subs.lattice.matrix[1, :],
                    film.lattice.matrix[2, :],
                ]
            )
        )
        film.modify_lattice(lmap)
        # print ("film",film)
        # print ("subs",subs)

        info["mismatch_u"] = mismatch_u
        info["mismatch_v"] = mismatch_v
        info["mismatch_angle"] = mismatch_angle
        info["area1"] = area1
        info["area2"] = area2
        info["film_sl"] = film
        info["subs_sl"] = subs
    except:
        pass
    return info


def get_hetero_type(A={}, B={}):
    stack = "na"
    int_type = "na"
    try:
        # if A['phi']>B['phi']:
        if A["scf_vbm"] - A["avg_max"] < B["scf_vbm"] - B["avg_max"]:
            stack = "BA"
        else:
            C = A
            D = B
            A = D
            B = C
            stack = "AB"
            # tmp=B
            # B=A
            # A=tmp
        vbm_a = A["scf_vbm"] - A["avg_max"]
        vbm_b = B["scf_vbm"] - B["avg_max"]
        cbm_a = A["scf_cbm"] - A["avg_max"]
        cbm_b = B["scf_cbm"] - B["avg_max"]
        #  print ('vbm_a,vbm_b,cbm_b,cbm_a',vbm_a,vbm_b,cbm_b,cbm_a)
        if vbm_a < vbm_b and vbm_b < cbm_b and cbm_b < cbm_a:
            int_type = "I"
        elif vbm_a < vbm_b and vbm_b < cbm_a and cbm_a < cbm_b:
            int_type = "II"
        elif vbm_a < cbm_a and cbm_a < vbm_b and vbm_b < cbm_b:
            int_type = "III"
    except:
        pass
    return int_type, stack


def get_hetero(film, substrate, seperation=3.0):

    # unique site coordinates in the substrate top layers
    coords_uniq_sub = substrate.cart_coords

    # unique site coordinates in the 2D material bottom layers
    coords_uniq_film = film.cart_coords

    substrate_top_z = max(substrate.cart_coords[:, 2])

    film_bottom = min(film.cart_coords[:, 2])

    # shift normal to the surface by 'seperation'
    sub_z = substrate.lattice.matrix[2, :]
    origin = np.array([0, 0, substrate_top_z])
    shift_normal = sub_z / np.linalg.norm(sub_z) * seperation
    # generate all possible interfaces, one for each combination of
    # unique substrate and unique 2d materials site in the layers .i.e
    # an interface structure for each parallel shift
    # interface = 2D material + substrate
    interface = substrate.copy()
    hetero_interfaces = []

    for site in film:
        new_coords = site.coords
        new_coords[2] = site.coords[2] - film_bottom
        new_coords = new_coords + origin + shift_normal
        interface.append(site.specie, new_coords, coords_are_cartesian=True)

    return interface


def get_direct_hetero(
    bulk_film="",
    bulk_subs="",
    film_miller=(1, 1, 1),
    film_subs=(0, 0, 1),
    seperation=3.0,
    film_layers=2,
    bulk_layers=1,
):
    bulk_film_cvn = SpacegroupAnalyzer(bulk_film).get_conventional_standard_structure()
    bulk_subs_cvn = SpacegroupAnalyzer(bulk_subs).get_conventional_standard_structure()
    slab_film = get_ase_surf(s=bulk_film_cvn, miller=film_miller, layers=film_layers)
    slab_subs = get_ase_surf(s=bulk_subs_cvn, miller=film_subs, layers=bulk_layers)
    hetero = get_hetero(slab_film, slab_subs, seperation=seperation)

    return hetero


def get_strt(jid=""):
    dat_2d = get_2d_dataset()
    for i in dat_2d:
        if i["jid"] == jid:
            return i["final_str"]


def make_2d_jid_strts(jid1="", jid2="", tol=3.0):
    mat1 = get_strt(jid1)
    mat2 = get_strt(jid2)
    rz1 = rangexyz(s=mat1)
    rz2 = rangexyz(s=mat2)
    x = get_direct_hetero(
        bulk_film=mat1,
        bulk_subs=mat2,
        film_miller=(0, 0, 1),
        film_subs=(0, 0, 1),
        film_layers=1,
        bulk_layers=1,
        seperation=tol,
    )
    return x
    # return x


if __name__ == "__main__":

    s = Structure.from_file("POSCAR-Al.vasp")
    x = get_direct_hetero(bulk_film=s, bulk_subs=s)
    print(Poscar(x))

    # x.to(fmt="poscar", filename="POSCAR-Al.vasp")
    x = make_2d_jid_strts(jid1="JVASP-664", jid2="JVASP-652")
    print(Poscar(x))

    s1 = Structure.from_file("POSCAR-Al.vasp")
    s2 = Structure.from_file("POSCAR-Al2O3.vasp")
    x = get_direct_hetero(bulk_film=s1, bulk_subs=s2)
    print(Poscar(x))



import os, builtins, io, pytest
from pymatgen.core.structure import Structure
from pymatgen.io.vasp.inputs import Poscar
from jarvis.lammps.jlammps import *
from jarvis.plane_defect.surface import surfer
from jarvis.point_defect.vacancy import vac_antisite_def_struct_gen


parameters = {
    "exec": "mpirun /cluster/bin/lmp_ctcms-14439-knc6 <in.elastic >out",
    "pair_coeff": "/users/knc6/Software/jarvis/jarvis/lammps/examples/Mishin-Ni-Al-2009.eam.alloy",
    "control_file": "/users/knc6/inelast.mod",
    "pair_style": "eam/alloy",
    "atom_style": "charge",
    "cluster": "pbs",
}


init = os.path.join(
    os.path.dirname(__file__),
    "..",
    "lammps",
    "examples",
    "Al03.eam.alloy_nist",
    "bulk@mp-134_fold",
    "mp-134",
    "init.mod",
)


dat = os.path.join(
    os.path.dirname(__file__),
    "..",
    "lammps",
    "examples",
    "Al03.eam.alloy_nist",
    "bulk@mp-134_fold",
    "mp-134",
    "data",
)
ff = os.path.join(
    os.path.dirname(__file__),
    "..",
    "lammps",
    "examples",
    "Al03.eam.alloy_nist",
    "bulk@mp-134_fold",
    "mp-134",
    "potential.mod",
)
log = os.path.join(
    os.path.dirname(__file__),
    "..",
    "lammps",
    "examples",
    "Al03.eam.alloy_nist",
    "bulk@mp-134_fold",
    "mp-134",
    "log.lammps",
)


def test_read_data():
    data = read_data(data=dat, ff=ff)
    assert len(data) == 1


def test_write_lammps_in():
    data = read_data(data=dat, ff=ff)
    fold = os.path.join(os.path.dirname(__file__), "tmp")
    if not os.path.exists(fold):
        os.makedirs(fold)
    file = os.path.join(os.path.dirname(__file__), fold, "tmp")
    write_lammps_in(
        structure=data,
        lammps_in=init,
        lammps_in1=ff,
        lammps_in2=file,
        parameters=parameters,
    )


def test_vac_antisite_def_struct_gen(tmpdir):
    data = read_data(data=dat, ff=ff)
    vacs = vac_antisite_def_struct_gen(struct=data, c_size=0, write_file=False)
    assert len(vacs) == 2


def sample_strt():
    data = read_data(data=dat, ff=ff)
    return data


# def test_get_get_phonopy_atoms():
#    s=sample_strt()
#    phn=get_phonopy_atoms(mat=s)
#    assert phn.symbols==['Al']


def test_write_lammps_data():
    success = False
    s = sample_strt()
    fold = dat = os.path.join(os.path.dirname(__file__), "tmp")
    if not os.path.exists(fold):
        os.makedirs(fold)
    file = dat = os.path.join(os.path.dirname(__file__), fold, "tmp")
    write_lammps_data(structure=s, file=file, write_tmp_file=False)
    success = True
    assert success == True


def test_analyz_loge():
    x = len(analyz_loge(log))
    assert x == 23


# def test_smart_surf():
#    s=sample_strt()
#    parameters = {'phonon_control_file':'/users/knc6/in.phonon','surf_control_file':'/users/knc6/inelast_nobox.mod','def_control_file':'/users/knc6/inelast_nobox.mod','json_dat':'/rk2/knc6/JARVIS-DFT/MDCS/all_mp.json','c_size':3,'vac_size':25,'surf_size':25,'phon_size':20,'cluster':'pbs','exec':'/users/knc6/Software/LAMMPS/lammps-master/src/lmp_serial <in.main >out','pair_style':'eam/alloy','pair_coeff':'../lammps/examples/Al03.eam.alloy','atom_style': 'charge' ,'control_file':'/users/knc6/inelast.mod'}
#    cwd=str(os.getcwd())
#    os.chdir('../lammps/examples/Al03.eam.alloy_nist/bulk@mp-134_fold')
#    sl,sh=smart_surf(strt=s,parameters=parameters)
#    print (sl,sh)
# test_smart_surf()



"""
To access data in JARVIS-API, you do not need to request an account
However, to upload your data, please request an account at https://jarvis.nist.gov/

Then install MDCS-api-tools by:

git clone https://github.com/knc6/MDCS-api-tools.git
cd MDCS-api-tools
python setup.py develop
"""

from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from monty.serialization import loadfn, dumpfn
import os, glob, json, sys

try:
    import mdcs
    from mdcs import curate, explore
except:
    print("Please install MDCS")
    pass
from monty.json import MontyEncoder, MontyDecoder
import numpy as np
from pymatgen.io.vasp.outputs import Vasprun
from pymatgen.core.structure import Structure

# import dicttoxml
from xml.etree.ElementTree import Element, SubElement, Comment, tostring, ElementTree

f = open("passwrd.txt", "r")  # path to your passowrd file
passd = f.read().splitlines()[0]
f.close()

user = "abc"  # your username
schema = "5a7f2a872887d500ab0c0d02"

# Converting data to xml
def data_json(
    energy="na",
    typ="na",
    formula="na",
    sgp="na",
    name="na",
    ref="na",
    func="na",
    elem="na",
    encut="na",
    kpoints="na",
    el_tens="na",
    KV="na",
    GV="na",
    m_eg="na",
    b_eg="na",
    op_eg="na",
    mbj_eg="na",
    en_arr="na",
    realx_arr="na",
    imagx_arr="na",
    realy_arr="na",
    imagy_arr="na",
    realz_arr="na",
    imagz_arr="na",
    men_arr="na",
    mrealx_arr="na",
    mimagx_arr="na",
    mrealy_arr="na",
    mimagy_arr="na",
    mrealz_arr="na",
    mimagz_arr="na",
    struct="na",
    other="na",
):
    """
    This is an example of how to upload jarvis data to the API
    Args:
        string key value pairs
    Returns:
          curate data at the API
    """

    top = Element("JARVIS-DFT")
    child = SubElement(top, "energy-ev")
    child.text = str(energy)
    child = SubElement(top, "formula")
    child.text = str(formula)
    child = SubElement(top, "space-group")
    child.text = str(sgp)
    child = SubElement(top, "JVID")
    child.text = str(name)
    child = SubElement(top, "reference")
    child.text = str(ref)
    child = SubElement(top, "calc_type")
    child.text = str(typ)
    child = SubElement(top, "functional")
    child.text = str(func)
    child = SubElement(top, "elements")
    child.text = str(elem)
    child = SubElement(top, "encut-ev")
    child.text = str(encut)
    child = SubElement(top, "kpoints")
    child.text = str(kpoints)
    child = SubElement(top, "elastic_tensor-gpa")
    child.text = str(el_tens)
    child = SubElement(top, "kv-gpa")
    child.text = str(KV)
    child = SubElement(top, "gv-gpa")
    child.text = str(GV)
    child = SubElement(top, "scf_eg-ev")
    child.text = str(m_eg)
    child = SubElement(top, "brill_eg-ev")
    child.text = str(b_eg)
    child = SubElement(top, "optics_eg-ev")
    child.text = str(op_eg)
    child = SubElement(top, "mbj-eg_ev")
    child.text = str(mbj_eg)
    child = SubElement(top, "opt_energy_arr-ev")
    child.text = str(en_arr)
    child = SubElement(top, "realx_arr")
    child.text = str(realx_arr)
    child = SubElement(top, "imagx_arr")
    child.text = str(imagx_arr)
    child = SubElement(top, "realy_arr")
    child.text = str(realy_arr)
    child = SubElement(top, "imagy_arr")
    child.text = str(imagy_arr)
    child = SubElement(top, "realz_arr")
    child.text = str(realz_arr)
    child = SubElement(top, "imagz_arr")
    child.text = str(imagz_arr)
    child = SubElement(top, "mbj_energy_arr-ev")
    child.text = str(men_arr)
    child = SubElement(top, "mbj_realx_arr")
    child.text = str(mrealx_arr)
    child = SubElement(top, "mbj_imagx_arr")
    child.text = str(mimagx_arr)
    child = SubElement(top, "mbj_realy_arr")
    child.text = str(mrealy_arr)
    child = SubElement(top, "mbj_imagy_arr")
    child.text = str(mimagy_arr)
    child = SubElement(top, "mbj_realz_arr")
    child.text = str(mrealz_arr)
    child = SubElement(top, "mbj_imagz_arr")
    child.text = str(mimagz_arr)
    child = SubElement(top, "structure")
    child.text = str(struct)
    child = SubElement(top, "other")
    child.text = str(other)
    filename = str(name) + str(".xml")
    ElementTree(top).write(filename)
    curate(
        filename, filename, schema, "https://jarvis.nist.gov/", user, passd, cert=False
    )


# Get a particular record
def get_record(file=""):
    """
   This is an example of how to get a particular jarvis-id document, say JVASP-1002.xml
   Args:
       file: name of the file
   Returns:
         data in json format
   """

    r = explore.select(
        "https://jarvis.nist.gov/", user, passd, cert=False, title=file, format="json"
    )
    return r


# Delete all record
def delete_all(file=""):
    """
    Caution, this deletes your documents
    """
    r = explore.select_all(
        "https://jarvis.nist.gov/", user, passd, cert=False, format="json"
    )
    for i in r:
        id = i["_id"]
        explore.delete(id, "https://jarvis.nist.gov/", user, passd, cert=False)


# Download 3D material data from https://figshare.com/articles/jdft_3d-7-7-2018_json/6815699
# Download 2D material data from https://figshare.com/articles/jdft_2d-7-7-2018_json/6815705
# Example of curating/uploading JARVIS-DFT data  is given below


def upload_sample_data():
    d = loadfn("jdft_3d-7-7-2018.json", cls=MontyDecoder)

    count = 0
    for i in d:
        filname = str(i["jid"]) + str(".xml")
        if not os.path.exists(filname):
            count = count + 1
            energy = str(i["fin_en"]) + str(",") + str(i["form_enp"])
            formula = str(i["final_str"].composition.reduced_formula)
            sgp = str(SpacegroupAnalyzer(i["final_str"]).get_spacegroup_symbol())
            name = str(i["jid"])
            print(name)
            ref = str(i["mpid"])
            func = str("OptB88vdW")
            elem = ""
            species = list(set(i["final_str"].species))
            for j in species:
                elem = str(elem) + str(j.symbol) + str("-")
            encut = str(i["encut"])
            kpoints = (
                str(i["kpoints"].kpts[0][0])
                + str("x")
                + str(i["kpoints"].kpts[0][1])
                + str("x")
                + str(i["kpoints"].kpts[0][2])
            )
            el_tens = str(i["elastic"])
            KV = str(i["kv"])
            GV = str(i["gv"])
            op_eg = str(i["op_gap"])
            mbj_eg = str(i["mbj_gap"])
            realx_arr = str(i["epsx"])
            mrealx_arr = str(i["mepsx"])
            realy_arr = str(i["epsy"])
            mrealy_arr = str(i["mepsy"])
            realz_arr = str(i["epsz"])
            mrealz_arr = str(i["mepsz"])
            typ = str("3D")
            other = str(
                "Citation: 1) DOI:10.1038/s41598-017-05402-0, 2) DOI: 10.1038/sdata.2018.82, 3) arXiv:1804.01033v2 "
            )
            struct = str(
                i["final_str"]
                .to(fmt="cif")
                .replace("# generated using pymatgen", "# JARVIS-DFT")
            )
            data_json(
                other=other,
                energy=energy,
                typ=typ,
                formula=formula,
                sgp=sgp,
                name=name,
                func=func,
                elem=elem,
                encut=encut,
                kpoints=kpoints,
                el_tens=el_tens,
                KV=KV,
                GV=GV,
                op_eg=op_eg,
                mbj_eg=mbj_eg,
                realx_arr=realx_arr,
                mrealx_arr=mrealx_arr,
                realy_arr=realy_arr,
                mrealy_arr=mrealy_arr,
                realz_arr=realz_arr,
                mrealz_arr=mrealz_arr,
                struct=struct,
            )


# Some examples

# data_json()

# x=get_record(file='JVASP-48137.xml')[0]['content']['JARVIS-DFT']['structure']

# print (Structure.from_str(x,fmt='cif'))
# delete_all()
# curate_data()
# delete_all()
# for k,v in r[0].items():
# for k,v in r[0]['content'].items():
#    print k



from __future__ import division, unicode_literals

"""
Surface and defect structure generator
See http://iopscience.iop.org/article/10.1088/1361-648X/aadaff/meta
and
See DOI: 10.1016/j.cpc.2015.03.015
"""


import argparse
import os
from pymatgen.analysis.defects.generators import VacancyGenerator
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
import glob
from pymatgen.io.vasp import Poscar
from pymatgen.core.structure import Structure
from pymatgen.io.ase import AseAtomsAdaptor
from monty.serialization import loadfn, MontyDecoder


def chempot_struct(element="Al"):
    chempot_json_file = str(os.path.join(os.path.dirname(__file__), "chem_pot.json"))
    chempot_json = loadfn(chempot_json_file, cls=MontyDecoder)
    id, struct = chempot_json[element]
    return id, struct


def vac_antisite_def_struct_gen(c_size=15, struct=None, write_file=True):
    """
    Vacancy, antisite generator

    Args:
         c_size: cell size
         struct: Structure object or
         mpid: materials project id
    Returns:
            def_str: defect structures in Poscar object format
    """
    def_str = []
    if struct == None:
        print("Provide structure")
    c_size = c_size
    prim_struct_sites = len(struct.sites)
    struct = SpacegroupAnalyzer(struct).get_conventional_standard_structure()
    dim1 = int((float(c_size) / float(max(abs(struct.lattice.matrix[0]))))) + 1
    dim2 = int(float(c_size) / float(max(abs(struct.lattice.matrix[1])))) + 1
    dim3 = int(float(c_size) / float(max(abs(struct.lattice.matrix[2])))) + 1
    cellmax = max(dim1, dim2, dim3)
    conv_struct_sites = len(struct.sites)
    conv_prim_rat = int(conv_struct_sites / prim_struct_sites)
    sc_scale = [dim1, dim2, dim3]
    print("sc_scale", sc_scale)

    tmp = struct.copy()
    tmp.make_supercell(sc_scale)
    sc_tmp = tmp  # Poscar(tmp).structure .make_supercell(list(sc_scale))
    scs = list(VacancyGenerator(struct))
    supercell = Poscar(sc_tmp)
    supercell.comment = str("bulk") + str("@") + str("cellmax") + str(cellmax)
    def_str.append(supercell)
    if write_file == True:
        supercell.write_file("POSCAR-" + str("bulk") + str(".vasp"))

    for i in range(len(scs)):
        sc = scs[i].generate_defect_structure(sc_scale)
        poscar = Poscar(sc)  # mpvis.get_poscar(sc)
        pmg_name = str(scs[i].name).split("_")
        sitespecie = pmg_name[1]
        mult = pmg_name[2].split("mult")[1]
        name = (
            str("vacancy_")
            + str(i + 1)
            + str("_mult-")
            + str(mult)
            + str("_sitespecie-")
            + str(sitespecie)
            + str("@cellmax")
            + str(cellmax)
        )
        poscar.comment = str(name)
        def_str.append(poscar)
        if write_file == True:
            filename = (
                str("POSCAR-")
                + str("vacancy_")
                + str(i + 1)
                + str("_mult-")
                + str(mult)
                + str("_sitespecie-")
                + str(sitespecie)
                + str(".vasp")
            )
            poscar.write_file(filename)

    return def_str


if __name__ == "__main__":
    pos = str(
        os.path.join(
            os.path.dirname(__file__),
            "../vasp/examples/SiOptb88/MAIN-MBJ-bulk@mp_149/POSCAR",
        )
    )
    mat = Structure.from_file(pos)
    x = vac_antisite_def_struct_gen(struct=mat, write_file=False)
    print(x)
    id, s = chempot_struct("Al")
    print(id, s)



"""
Integration of JARVIS-FF/DFT/ML .
"""
import os
from pathlib import Path

try:

    dir = os.path.join(os.path.dirname(__file__))
    #dir = os.path.dirname(os.path.realpath('__file__'))
    env = os.path.join(dir, "env_variables")

    f = open(env, "r")
    lines = f.read().splitlines()
    f.close()
    for i in lines:
        tmp = i.split("=")
        if len(tmp) > 1:
            os.environ[tmp[0]] = tmp[1]
except:
    pass



"""
Simple ML models for classifcation and regression, designed for educational purposes only
__author__: = Kamal Choudhary
"""

import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.ensemble import (
    RandomForestRegressor,
    GradientBoostingRegressor,
    AdaBoostRegressor,
)
from sklearn.svm import SVR
from sklearn.linear_model import Lasso, LinearRegression, LogisticRegression
from sklearn.kernel_ridge import KernelRidge
from sklearn.neural_network import MLPRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.multiclass import OneVsRestClassifier
from sklearn.datasets import load_boston
from sklearn.metrics import roc_curve, auc
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import label_binarize
from sklearn.preprocessing import StandardScaler
from sklearn.feature_selection import VarianceThreshold
from sklearn.pipeline import Pipeline
from sklearn.neural_network import MLPClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.ensemble import (
    RandomForestClassifier,
    AdaBoostClassifier,
    GradientBoostingClassifier,
)
from sklearn.svm import SVC
from pymatgen.core.structure import Structure
from jarvis.sklearn.build_models import *
from jarvis.sklearn.build_models import jdata
from jarvis.db.static.explore_db import get_3d_dataset, get_2d_dataset, get_ml_dataset

simple_regr_models = [
    GaussianProcessRegressor(),
    RandomForestRegressor(),
    GradientBoostingRegressor(),
    AdaBoostRegressor(),
    SVR(),
    Lasso(),
    LinearRegression(),
    KernelRidge(),
    MLPRegressor(),
    DecisionTreeRegressor(),
]

simple_class_models = [
    DecisionTreeClassifier(),
    MLPClassifier(),
    GradientBoostingClassifier(),
    KNeighborsClassifier(),
    GaussianProcessClassifier(),
    RandomForestClassifier(),
    AdaBoostClassifier(),
    SVC(),
]


def read_vasp_structure(filename=""):
    """
   Read VASP POSCAR, .cif file
   Args:
       filename:string
   Returns:
       s: pymatgen.structure object
   """

    s = Structure.from_file(filename)
    return s


def simple_regression(
    X=[], Y=[], plot=False, simple_models=simple_regr_models, preprocess=True
):
    """
  Quickly train simple regression models without hyperparameter-optimization
  Args:
      X: input features
      Y: Target data
      plot: whether to make a parity plot with ML models
      preprocess: whether to apply standard preprocessing techniques
  Returns:
  """

    X_train, X_test, y_train, y_test, jid_train, jid_test = train_test_split(
        X, Y, jid, random_state=1, test_size=0.1
    )
    for i in simple_models:
        pipe = Pipeline(
            [
                ("stdscal", StandardScaler()),
                ("vart", VarianceThreshold(1e-4)),
                ("est", i),
            ]
        )
        if preprocess == True:
            model = pipe
        else:
            model = i
        model.fit(X_train, y_train)
        pred = model.predict(X_test)
        reg_sc = regr_scores(y_test, pred)
        if plot == True:
            plt.plot(
                reg_sc["pred"], reg_sc["test"], ".", label=str(type(i).__name__)[0:4]
            )
        print(type(i).__name__, round(reg_sc["mae"], 3), round(reg_sc["rmse"], 3))
    if plot == True:
        plt.legend()
        plt.xlabel("DFT")
        plt.ylabel("ML")


def binary_class_dat(X=[], Y=[], tol=0.1):
    """
    Categorize a continous dataset in 1/0 with a threshold "tol"
    """

    Y1 = []
    for i, j in zip(X, Y):
        if j >= tol:
            Y1.append(1)
        else:
            Y1.append(0)
    return X, Y1


def bar_plot(x=[], interval=10):
    """
  Plot bar plot to understand the range of data
  """
    # interval = # abs(max(x)-min(x))/len(x)
    hist, bins = np.histogram(
        x, bins=np.arange(min(x), max(x), interval), density=False
    )
    return bins[:-1], hist


def classify_roc_ml(
    X=[],
    y=[],
    classes=[0, 1, 2],
    names=["High val", "Low val", ""],
    n_plot=1,
    method="",
    preprocess=True,
    plot=False,
):
    """
    Classifcation module for ROC curve for upto three classes, can be expanded in more classes as well
    Args:
        X: input feature vectors
        y: target data obtained from binary_class_dat
        classes: dummy classes
        names: name holders for the target data
        method: ML method
        preprocess: whether to apply standard preprocessing techniques
        plt: whether to plot the ROC curve
    """
    if plot == True:
        plt.close()
        plt.rcParams.update({"font.size": 22})
        plt.figure(figsize=(12, 8))

    y = label_binarize(y, classes=classes)
    n_classes = y.shape[1]

    pipe = Pipeline(
        [
            ("stdscal", StandardScaler()),
            ("vart", VarianceThreshold(1e-4)),
            ("est", method),
        ]
    )
    if preprocess == True:
        model = pipe
    else:
        model = method
    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.1, random_state=0
    )
    classifier = OneVsRestClassifier(model)
    if hasattr(model, "decision_function"):
        y_score = classifier.fit(X_train, y_train).decision_function(X_test)
    else:
        y_score = classifier.fit(X_train, y_train).predict_proba(X_test)

    lw = 3

    fpr = dict()
    tpr = dict()
    roc_auc = dict()
    for i in range(n_classes):
        fpr[i], tpr[i], _ = roc_curve(y_test[:, i], y_score[:, i])
        roc_auc[i] = auc(fpr[i], tpr[i])
    colors = ["blue", "red", "green"]
    count = 0
    for i, color, name in zip(range(n_classes), colors, names):

        if name != "":
            if count < n_plot:
                count = count + 1
                if plot == True:
                    plt.plot(
                        fpr[i],
                        tpr[i],
                        color=color,
                        lw=lw,
                        label="ROC  {0} (area = {1:0.2f})" "".format(name, roc_auc[i]),
                    )
    if plot == True:
        plt.plot([0, 1], [0, 1], "k--", lw=lw)
        plt.xlim([-0.05, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel("False Positive Rate")
        plt.ylabel("True Positive Rate")
        plt.legend(loc="lower right")
        plt.show()
    model = classifier.fit(X, y)
    return model, roc_auc


def simple_classification(plot=True, simple_models=simple_class_models):
    """
    Quickly train some of the classifcation algorithms available in scikit-learn
    """

    for i in simple_models:
        m, r = classify_roc_ml(X=X_class, y=Y_class, method=i, plot=plot)
        print(type(i).__name__, r[0])


def simple_GB_confidence_interval(X=[], Y=[]):
    """
 From https://towardsdatascience.com/how-to-generate-prediction-intervals-with-scikit-learn-and-python-ab3899f992ed

 """
    LOWER_ALPHA = 0.1
    UPPER_ALPHA = 0.9
    # Each model has to be separate
    lower_model = GradientBoostingRegressor(loss="quantile", alpha=LOWER_ALPHA)
    # The mid model will use the default loss
    mid_model = GradientBoostingRegressor(loss="ls")
    upper_model = GradientBoostingRegressor(loss="quantile", alpha=UPPER_ALPHA)

    models = [lower_model, mid_model, upper_model]
    simple_regression(X=X, Y=Y, simple_models=models)


if __name__ == "__main__":
    X, Y, jid = jdata(prop="exfoliation_en")
    simple_GB_confidence_interval(X, Y)
    simple_regression(X, Y)
    X_class, Y_class = binary_class_dat(X=X, Y=Y, tol=100)
    simple_classification()



from __future__ import unicode_literals

# source ~/anaconda2/envs/my_boltz2/bin/activate my_boltz2
import matplotlib.pyplot as plt

plt.switch_backend("Agg")
import unittest
import json, os
import warnings
from monty.serialization import loadfn, dumpfn, MontyEncoder, MontyDecoder
from pymatgen.io.vasp import Vasprun
from interruptingcow import timeout
from pymatgen.electronic_structure.boltztrap2 import (
    BandstructureLoader,
    VasprunLoader,
    BztInterpolator,
    BztTransportProperties,
    BztPlotter,
    merge_up_down_doses,
)


from pymatgen.electronic_structure.core import Spin, OrbitalType

import numpy as np
from monty.serialization import loadfn


vrunfile = "pymatgen-master/test_files/boltztrap2/vasprun.xml"
vrunfile1 = "/users/knc6/Software/JARVIS-TOOLS/jarvis/jarvis/vasp/examples/SiOptb88/MAIN-RELAX-bulk@mp_149/vasprun.xml"
vrunfile2 = "/users/knc6/Software/JARVIS-TOOLS/jarvis/jarvis/vasp/examples/SiOptb88/MAIN-MBJ-bulk@mp_149/vasprun.xml"
vrunfile3 = "/users/knc6/Software/JARVIS-TOOLS/jarvis/jarvis/vasp/examples/SiOptb88/MAIN-OPTICS-bulk@mp_149/vasprun.xml"


@timeout(3600)
def nonspin_boltz(vrunfile="", k_latt=1.0, write_json=False):
    fname = vrunfile.replace("vasprun.xml", "boltz2data.json")
    if not os.path.isfile(fname):
        vrun = Vasprun(vrunfile, parse_projected_eigen=True)
        loader = VasprunLoader().from_file(vrunfile)
        temp_r = np.array([300, 400, 500, 600, 700, 800])
        doping = np.array([0, 10 ** 18, 10 ** 19, 10 ** 20, 10 ** 21, 10 ** 22])
        bztInterp = BztInterpolator(loader, lpfac=2)
        bztTransp = BztTransportProperties(bztInterp, temp_r=temp_r, doping=doping)
        xx = bztTransp.compute_properties_doping(doping=doping)
        # 4 temps, 2 doping
        Conductivity_doping = bztTransp.Conductivity_doping
        Seebeck_doping = bztTransp.Seebeck_doping
        Kappa_doping = bztTransp.Kappa_doping
        Effective_mass_doping = bztTransp.Effective_mass_doping
        Power_Factor_doping = bztTransp.Power_Factor_doping
        mu_r_eV = bztTransp.mu_r_eV
        info = {}
        info["mu_r_eV"] = mu_r_eV
        info["temp_r"] = temp_r
        info["doping"] = doping
        info["Conductivity_doping"] = Conductivity_doping
        info["Seebeck_doping"] = Seebeck_doping
        info["Kappa_doping"] = Kappa_doping
        info["Effective_mass_doping"] = Effective_mass_doping
        info["Power_Factor_doping"] = Power_Factor_doping

        info["Conductivity_mu"] = bztTransp.Conductivity_mu
        info["Seebeck_mu"] = bztTransp.Seebeck_mu
        info["Kappa_mu"] = bztTransp.Kappa_mu
        info["Power_Factor_mu"] = bztTransp.Power_Factor_mu
        info["Effective_mass_mu"] = bztTransp.Effective_mass_mu
        info["Hall_carrier_conc_trace_mu"] = bztTransp.Hall_carrier_conc_trace_mu

        zt = []
        temp_zt = []
        for i, ii in enumerate(info["temp_r"]):
            for j, k in zip(info["Power_Factor_mu"][i], info["Kappa_mu"][i]):
                temp_zt.append(0.001 * j * ii / (k + k_latt))
            zt.append(temp_zt)
            temp_zt = []
        zt = np.array(zt)
        info["zt_mu"] = zt
        if write_json == True:
            f = open(fname, "w")
            f.write(json.dumps(info, cls=MontyEncoder))
            f.close()

        return info
    else:
        print("File exists")
    # print (bztTransp.Effective_mass_mu[0][0])
    # bztPlotter = BztPlotter(bztTransp,bztInterp)
    # fig = bztPlotter.plot_props('S', 'mu', 'temp', temps=[300, 500])
    # fig.savefig('first_boltz2.png')
    # fig.close()


@timeout(3600)
def spin_boltz(vrunfile="", spin=1, k_latt=1.0, write_json=True):
    fname = vrunfile.replace("vasprun.xml", "boltz2data.json")
    if not os.path.isfile(fname):
        kp = vrunfile.replace("vasprun.xml", "KPOINTS")
        v = Vasprun(vrunfile)
        nelect = v.parameters["NELECT"]
        bs = v.get_band_structure(kp, line_mode=False)
        # doping=10.**np.arange(20,22)
        temp_r = np.array([300, 400, 500, 600, 700, 800])
        doping = np.array([0, 10 ** 18, 10 ** 19, 10 ** 20, 10 ** 21, 10 ** 22])
        loader = BandstructureLoader(bs, v.structures[-1], spin=spin, nelect=nelect)
        bztInterp = BztInterpolator(loader, lpfac=2, curvature=True)
        bztTransp = BztTransportProperties(bztInterp, doping=doping, temp_r=temp_r)
        xx = bztTransp.compute_properties_doping(doping=doping)
        # 4 temps, 2 doping
        Conductivity_doping = bztTransp.Conductivity_doping
        Seebeck_doping = bztTransp.Seebeck_doping
        Kappa_doping = bztTransp.Kappa_doping
        Effective_mass_doping = bztTransp.Effective_mass_doping
        Power_Factor_doping = bztTransp.Power_Factor_doping
        mu_r_eV = bztTransp.mu_r_eV

        info = {}
        info["mu_r_eV"] = mu_r_eV
        info["temp_r"] = temp_r
        info["doping"] = doping
        info["Conductivity_doping"] = Conductivity_doping
        info["Seebeck_doping"] = Seebeck_doping
        info["Kappa_doping"] = Kappa_doping
        info["Effective_mass_doping"] = Effective_mass_doping
        info["Power_Factor_doping"] = Power_Factor_doping

        info["Conductivity_mu"] = bztTransp.Conductivity_mu
        info["Seebeck_mu"] = bztTransp.Seebeck_mu
        info["Kappa_mu"] = bztTransp.Kappa_mu
        info["Power_Factor_mu"] = bztTransp.Power_Factor_mu
        info["Effective_mass_mu"] = bztTransp.Effective_mass_mu
        info["Hall_carrier_conc_trace_mu"] = bztTransp.Hall_carrier_conc_trace_mu
        zt = []
        temp_zt = []
        for i, ii in enumerate(info["temp_r"]):
            for j, k in zip(info["Power_Factor_mu"][i], info["Kappa_mu"][i]):
                temp_zt.append(0.001 * j * ii / (k + k_latt))
            zt.append(temp_zt)
            temp_zt = []
        zt = np.array(zt)
        info["zt_mu"] = zt
        if write_json == True:
            f = open(fname, "w")
            f.write(json.dumps(info, cls=MontyEncoder))
            f.close()
        return info
    else:
        print("File exists")


def run_dir(
    dir="/users/knc6/Software/JARVIS-TOOLS/jarvis/jarvis/vasp/examples/SiOptb88"
):
    for i in os.listdir(dir):
        if "json" in i:
            if "MAIN-RELAX" in i:
                try:
                    mvrun = (
                        str(dir)
                        + str("/")
                        + str(i.split(".json")[0])
                        + str("/vasprun.xml")
                    )
                    info = spin_boltz(mvrun)
                    print("mvrun", mvrun, info)
                except:
                    print("skpiing")
                    pass
            if "MAIN-MBJ" in i:
                try:
                    mbjvrun = (
                        str(dir)
                        + str("/")
                        + str(i.split(".json")[0])
                        + str("/vasprun.xml")
                    )
                    info = spin_boltz(mbjvrun)
                    print("mbjvrun", mbjvrun, info)
                except:
                    print("skpiing")
                    pass
            if "MAIN-OPTICS" in i:
                try:
                    mopvrun = (
                        str(dir)
                        + str("/")
                        + str(i.split(".json")[0])
                        + str("/vasprun.xml")
                    )
                    info = spin_boltz(mopvrun)
                    print("mopvrun", mopvrun, info)
                except:
                    print("skpiing")
                    pass


def get_specific_data(my_type="p", my_dop=10 ** 20, my_temp=600, bfile=""):
    dop = "na"
    temp = "na"
    d = loadfn(bfile, cls=MontyDecoder)
    for i, ii in enumerate(d["temp_r"]):
        if ii == my_temp:
            temp = i
    for j, jj in enumerate(d["doping"]):
        if jj == my_dop:
            dop = j
    if dop != "na" and temp != "na":
        seeb = max(d["Seebeck_doping"][my_type][temp][dop].flatten())
        cond = max(d["Conductivity_doping"][my_type][temp][dop].flatten())
        kapp = max(d["Kappa_doping"][my_type][temp][dop].flatten())
        pf = max(d["Power_Factor_doping"][my_type][temp][dop].flatten())
        eff_mass = max(d["Effective_mass_doping"][my_type][temp][dop].flatten())
        return seeb, cond, kapp, pf, eff_mass
    print("Selection does not exist")


def plot_prop_mu(prop="Seebeck_mu", bfile="", temp=300, filename="seebplot.png"):

    props = [
        "Conductivity_mu",
        "Seebeck_mu",
        "Kappa_mu",
        "Effective_mass_mu",
        "Power_Factor_mu",
        "Carrier_conc_mu",
        "Hall_carrier_conc_trace_mu",
        "zt_mu",
    ]
    props_lbl = [
        "Conductivity",
        "Seebeck",
        "$K_{el}$",
        "Effective mass",
        "Power Factor",
        "Carrier concentration",
        "Hall carrier conc.",
        "ZT",
    ]
    props_unit = (
        r"$(\mathrm{kS\,m^{-1}})$",
        r"($\mu$V/K)",
        r"$(W / (m \cdot K))$",
        r"$(m_e)$",
        r"$( mW / (m\cdot K^2)$",
        r"$(cm^{-3})$",
        r"$(cm^{-3})$",
        "",
    )
    ylbl = "na"
    for i, j, k in zip(props, props_lbl, props_unit):
        if i == prop:
            ylbl = j + " " + k
    print("ylbl", ylbl)
    d = loadfn(bfile, cls=MontyDecoder)
    itemp = "na"
    for i, ii in enumerate(d["temp_r"]):
        if ii == temp:
            itemp = i
    data = d[prop][itemp]
    dat_x = []
    dat_y = []
    dat_z = []
    for i in data:
        dat, vecs = np.linalg.eig(i)
        dat_x.append(dat[0])
        dat_y.append(dat[1])
        dat_z.append(dat[2])
    plt.close()
    plt.rcParams.update({"font.size": 22})
    fig = plt.figure(figsize=(10, 8))
    plt.plot(d["mu_r_eV"], dat_x, label="x")
    plt.plot(d["mu_r_eV"], dat_y, label="y")
    plt.plot(d["mu_r_eV"], dat_z, label="z")
    # plt.axhline(y=0.0, color='k', linestyle='-')
    plt.ylabel(ylbl)
    plt.xlim([-3, 3])
    plt.xlabel(r"$\mu$ (eV)")
    plt.legend(loc="best")
    plt.tight_layout()
    plt.savefig(filename)
    plt.close()


if __name__ == "__main__":
    path = path = str(
        os.path.join(os.path.dirname(__file__), "../vasp/examples/SiOptb88")
    )
    run_dir(path)



import os, builtins, io, pytest
from pymatgen.core.structure import Structure
from pymatgen.io.vasp.inputs import Poscar

# from jarvis.lammps.jlammps import *
from jarvis.plane_defect.surface import surfer
from jarvis.point_defect.vacancy import vac_antisite_def_struct_gen

from jarvis.heterostruct.interface import *

pos = os.path.join(
    os.path.dirname(__file__), "..", "vasp", "examples", "SiOptb88", "POSCAR"
)


def test_get_ase_surf():
    s = Structure.from_file(pos)
    surf = get_ase_surf(s)
    assert (surf.lattice.matrix[2][2]) == 24.78934299449321


def test_mismatch_strts():
    s = Structure.from_file(pos)
    surf = get_ase_surf(s)
    info = mismatch_strts(film=surf, subs=surf)
    assert info["area1"] == 12.95010493389544


def test_get_hetero_type():
    A = {}
    B = {}
    A["scf_vbm"] = 1.0
    A["scf_cbm"] = 3.0
    A["avg_max"] = 2.5

    B["scf_vbm"] = 1.5
    B["scf_cbm"] = 2.0
    B["avg_max"] = 4.5
    int_type, stack = get_hetero_type(A=A, B=B)
    assert (int_type, stack) == ("III", "AB")


def test_get_direct_hetero():
    s = Structure.from_file(pos)
    x = get_direct_hetero(bulk_film=s, bulk_subs=s)
    print(len(x))
    assert len(x) == 24



from __future__ import division, unicode_literals

"""
Surface and defect structure generator
See http://iopscience.iop.org/article/10.1088/1361-648X/aadaff/meta
and
See DOI: 10.1016/j.cpc.2015.03.015
"""


import argparse
import os
from pymatgen.io.vasp import Poscar
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
import glob
from pymatgen.io.vasp import Poscar
from pymatgen.core.surface import Slab, SlabGenerator
from pymatgen.core.structure import Structure
from pymatgen.io.ase import AseAtomsAdaptor
from ase.lattice.surface import surface
from pymatgen.core.surface import (
    Slab,
    SlabGenerator,
    generate_all_slabs,
    get_symmetrically_distinct_miller_indices,
)


def pmg_surfer(vacuum=15, mat=None, max_index=1, min_slab_size=15, write_file=True):
    """
    Pymatgen surface builder for a Poscar

    Args:
        vacuum: vacuum region
        mat: Structure object
        max_index: maximum miller index
        min_slab_size: minimum slab size

    Returns:
           structures: list of surface Structure objects
    """

    if mat == None:
        print("Provide structure")

    sg_mat = SpacegroupAnalyzer(mat)
    mat_cvn = sg_mat.get_conventional_standard_structure()
    mat_cvn.sort()
    indices = get_symmetrically_distinct_miller_indices(mat_cvn, max_index)

    structures = []
    pos = Poscar(mat_cvn)
    try:
        pos.comment = (
            str("sbulk")
            + str("@")
            + str("vac")
            + str(vacuum)
            + str("@")
            + str("size")
            + str(min_slab_size)
        )
    except:
        pass
    structures.append(pos)
    if write_file == True:
        mat_cvn.to(fmt="poscar", filename=str("POSCAR-") + str("cvn") + str(".vasp"))
    for i in indices:
        slab = SlabGenerator(
            initial_structure=mat_cvn,
            miller_index=i,
            min_slab_size=min_slab_size,
            min_vacuum_size=vacuum,
            lll_reduce=False,
            center_slab=True,
            primitive=False,
        ).get_slab()
        normal_slab = slab.get_orthogonal_c_slab()
        slab_pymatgen = Poscar(normal_slab).structure
        xy_size = min_slab_size
        dim1 = (
            int((float(xy_size) / float(max(abs(slab_pymatgen.lattice.matrix[0]))))) + 1
        )
        dim2 = (
            int(float(xy_size) / float(max(abs(slab_pymatgen.lattice.matrix[1])))) + 1
        )
        slab_pymatgen.make_supercell([dim1, dim2, 1])
        slab_pymatgen.sort()
        surf_name = "_".join(map(str, i))
        pos = Poscar(slab_pymatgen)
        try:
            pos.comment = (
                str("Surf-")
                + str(surf_name)
                + str("@")
                + str("vac")
                + str(vacuum)
                + str("@")
                + str("size")
                + str(min_slab_size)
            )
        except:
            pass
        if write_file == True:
            pos.write_file(
                filename=str("POSCAR-") + str("Surf-") + str(surf_name) + str(".vasp")
            )
        structures.append(pos)

    return structures


def surfer(vacuum=15, layers=2, mat=None, max_index=1, write_file=True):
    """
    ASE surface bulder

    Args:
        vacuum: vacuum region
        mat: Structure object
        max_index: maximum miller index
        min_slab_size: minimum slab size

    Returns:
           structures: list of surface Structure objects
    """

    if mat == None:
        print("Provide structure")

    sg_mat = SpacegroupAnalyzer(mat)
    mat_cvn = sg_mat.get_conventional_standard_structure()
    mat_cvn.sort()
    indices = get_symmetrically_distinct_miller_indices(mat_cvn, max_index)
    ase_atoms = AseAtomsAdaptor().get_atoms(mat_cvn)

    structures = []
    pos = Poscar(mat_cvn)
    try:
        pos.comment = (
            str("sbulk")
            + str("@")
            + str("vac")
            + str(vacuum)
            + str("@")
            + str("layers")
            + str(layers)
        )
    except:
        pass
    structures.append(pos)
    if write_file == True:
        mat_cvn.to(fmt="poscar", filename=str("POSCAR-") + str("cvn") + str(".vasp"))
    for i in indices:
        ase_slab = surface(ase_atoms, i, layers)
        ase_slab.center(vacuum=vacuum, axis=2)
        slab_pymatgen = AseAtomsAdaptor().get_structure(ase_slab)
        slab_pymatgen.sort()
        surf_name = "_".join(map(str, i))
        pos = Poscar(slab_pymatgen)
        try:
            pos.comment = (
                str("Surf-")
                + str(surf_name)
                + str("@")
                + str("vac")
                + str(vacuum)
                + str("@")
                + str("layers")
                + str(layers)
            )
        except:
            pass
        if write_file == True:
            pos.write_file(
                filename=str("POSCAR-") + str("Surf-") + str(surf_name) + str(".vasp")
            )
        structures.append(pos)

    return structures


if __name__ == "__main__":
    pos = str(
        os.path.join(
            os.path.dirname(__file__),
            "../vasp/examples/SiOptb88/MAIN-MBJ-bulk@mp_149/POSCAR",
        )
    )
    mat = Structure.from_file(pos)
    x = pmg_surfer(mat=mat, write_file=False)
    print(x)
    y = surfer(mat=mat, write_file=False)



"""
Classical Force-field Inspired Descriptors (CFID)
Find details in:
https://journals.aps.org/prmaterials/abstract/10.1103/PhysRevMaterials.2.083801
"""
from __future__ import unicode_literals, print_function
import matplotlib.pyplot as plt

plt.switch_backend("agg")
from collections import defaultdict
import itertools
from scipy.stats import gaussian_kde
from math import pi
from operator import itemgetter
from pymatgen.core.structure import Structure
from pymatgen.io.vasp.outputs import Vasprun

# from pymatgen.analysis.structure_analyzer import VoronoiCoordFinder
import collections, math, os

# from pymatgen.analysis.defects.point_defects import ValenceIonicRadiusEvaluator
import numpy as np
from math import log
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from monty.json import MontyEncoder, MontyDecoder
from pymatgen.core.periodic_table import Element
import json, sys
from pymatgen.core.structure import Structure
import numpy as np
import xml.etree.ElementTree as ET
from monty.serialization import loadfn
from pandas import DataFrame as pd
from pymatgen.core.lattice import Lattice
from pymatgen.core import Composition
from interruptingcow import timeout

# Note time limit for angular part is hardcoded

el_chrg_json = str(os.path.join(os.path.dirname(__file__), "element_charge.json"))
el_chem_json = str(os.path.join(os.path.dirname(__file__), "Elements.json"))
timelimit = 300  # 5 minutes


def get_effective_structure(s=None, tol=8.0):
    """
  Check if there is vacuum, if so get actual size of the structure
  and the add vaccum of size tol to make sure structures
  are independent of user defined vacuum

  Args:
       s: Structure object
       tol: vacuum tolerance
  Returns:
         s: re-structure structure with tol vacuum
  """
    coords = s.cart_coords
    range_x = abs(max(coords[:, 0]) - min(coords[:, 0]))
    range_y = abs(max(coords[:, 1]) - min(coords[:, 1]))
    range_z = abs(max(coords[:, 2]) - min(coords[:, 2]))
    a = abs(s.lattice.a)  # matrix[0][0]
    b = abs(s.lattice.b)  # matrix[1][1]
    c = abs(s.lattice.c)  # matrix[2][2]
    if abs(a - range_x) > tol:
        a = range_x + tol
        print("Find vaccum in x-direction")
    if abs(b - range_y) > tol:
        b = range_y + tol
        print("Find vaccum in y-direction")
    if abs(c - range_z) > tol:
        c = range_z + tol
        print("Find vaccum in z-direction")
    arr = Lattice(
        [
            [s.lattice.matrix[0][0], s.lattice.matrix[0][1], s.lattice.matrix[0][2]],
            [s.lattice.matrix[1][0], s.lattice.matrix[1][1], s.lattice.matrix[1][2]],
            [s.lattice.matrix[2][0], s.lattice.matrix[2][1], s.lattice.matrix[2][2]],
        ]
    )
    s = Structure(arr, s.species, coords, coords_are_cartesian=True)
    return s


def el_combs(s=[]):
    """
  Get element combinations for a Structure object

  Args:
      s: Structure object
  Returns: 
         comb: combinations
    """
    symb = s.symbol_set
    tmp = map("-".join, itertools.product(symb, repeat=2))
    comb = list(set([str("-".join(sorted(i.split("-")))) for i in tmp]))
    return comb


def flatten_out(arr=[], tol=0.1):
    """
    Determine first cut-off
 
    Args:
         arr: array
         tol: toelrance
    Return:
          rcut: cut-off for a given tolerance tol, 
          because sometimes RDF peaks could be very close
    """

    rcut_buffer = tol
    io1 = 0
    io2 = 1
    io3 = 2
    delta = arr[io2] - arr[io1]

    while delta < rcut_buffer and io3 < len(arr):
        io1 = io1 + 1
        io2 = io2 + 1
        io3 = io3 + 1
        delta = arr[io2] - arr[io1]
    rcut = (arr[io2] + arr[io1]) / float(2.0)
    return rcut


def smooth_kde(x, y):
    """
    For making smooth distributions
    """
    denn = gaussian_kde(y)
    denn.covariance_factor = lambda: 0.25
    denn._compute_covariance()
    xs = np.linspace(0, max(x), 100)
    kde = denn(xs)
    return kde


def get_prdf(
    s=None, cutoff=10.0, intvl=0.1, plot_prdf=False, std_tol=0.25, filename="prdf.png"
):
    """
    Get partial radial distribution function

    Args:
         s: Structure object
         cutoff: maximum cutoff in Angstrom
         intvl: bin-size
         plot_prdf: whether to plot PRDF
         std_tol: when calculating the dihedral angles, the code got stuck for some N-containing compounds
         so we averaged all the first neghbors and added a fraction of the standard deviation as a tolerance
         this new tolerance is used for dihedral part only, while angular and RDF part remains intact
         filename: if plotted the name of the output file
    Returns:
          max-cutoff to ensure all the element-combinations are included
    """

    neighbors_lst = s.get_all_neighbors(cutoff)
    comb = el_combs(
        s=s
    )  # [str(i[0])+str('-')+str(i[1]) for i in list(itertools.product(sps, repeat=2))]
    info = {}
    for c in comb:
        for i, ii in enumerate(neighbors_lst):
            for j in ii:
                comb1 = str(s[i].specie) + str("-") + str(j[0].specie)
                comb2 = str(j[0].specie) + str("-") + str(s[i].specie)
                if comb1 == c or comb2 == c:
                    info.setdefault(c, []).append(j[1])
    plt.rcParams.update({"font.size": 22})
    for i in info.items():
        i[1].sort()
        dist_hist, dist_bins = np.histogram(
            i[1], bins=np.arange(0, cutoff + intvl, intvl), density=False
        )
        shell_vol = (
            4.0 / 3.0 * pi * (np.power(dist_bins[1:], 3) - np.power(dist_bins[:-1], 3))
        )
        number_density = s.num_sites / s.volume
        rdf = dist_hist / shell_vol / number_density / len(neighbors_lst)
        if plot_prdf == True:
            plt.plot(dist_bins[:-1], rdf, label=i[0], linewidth=2)

    if plot_prdf == True:

        plt.legend(prop={"size": 26})
        plt.xlabel("r ($\AA$)")
        plt.ylabel("g(r)")
        plt.ylim(ymin=0)
        plt.tight_layout()
        return plt
        # plt.savefig(filename)
        # plt.close()

    cut_off = {}

    for i, j in info.items():
        cut_off[i] = flatten_out(arr=j, tol=0.1)
    vals = np.array([float(i) for i in cut_off.values()])
    # print (cut_off.items(),' fwfwf',np.mean(vals)+0.25*np.std(vals))
    return (
        max(cut_off.items(), key=itemgetter(1))[1],
        np.mean(vals) + 0.25 * np.std(vals),
    )


def get_rdf(s=None, cutoff=10.0, intvl=0.1):
    """
    Get total radial distribution function

    Args:
         s: Structure object
         cutoff: maximum distance for binning
         intvl: bin-size
    Returns:
           bins, RDF, bond-order distribution
    """

    neighbors_lst = s.get_all_neighbors(cutoff)
    all_distances = np.concatenate(
        tuple(map(lambda x: [itemgetter(1)(e) for e in x], neighbors_lst))
    )
    rdf_dict = {}
    dist_hist, dist_bins = np.histogram(
        all_distances, bins=np.arange(0, cutoff + intvl, intvl), density=False
    )  # equivalent to bon-order
    shell_vol = (
        4.0 / 3.0 * pi * (np.power(dist_bins[1:], 3) - np.power(dist_bins[:-1], 3))
    )
    number_density = s.num_sites / s.volume
    rdf = dist_hist / shell_vol / number_density / len(neighbors_lst)
    return (
        dist_bins[:-1],
        [round(i, 4) for i in rdf],
        dist_hist / float(len(s)),
    )  # [{'distances': dist_bins[:-1], 'distribution': rdf}]


def get_dist_cutoffs(s="", max_cut=5.0):
    """
    Helper function to get dufferent type of distance cut-offs
    Args:
        s: Structure object
    Returns:
           rcut: max-cutoff to ensure all the element-combinations are included, used in calculating angluar distribution upto first neighbor
           rcut1: decide first cut-off based on total RDF and a buffer (previously used in dihedrals, but not used now in the code)
           rcut2: second neighbor cut-off
           rcut_dihed: specialized cut-off for dihedrals to avaoid large bonds such as N-N, uses average bond-distance and standard deviations
    """

    x, y, z = get_rdf(s)
    arr = []
    for i, j in zip(x, z):
        if j > 0.0:
            arr.append(i)
    rcut_buffer = 0.11
    io1 = 0
    io2 = 1
    io3 = 2
    delta = arr[io2] - arr[io1]
    while delta < rcut_buffer and arr[io2] < max_cut:
        io1 = io1 + 1
        io2 = io2 + 1
        io3 = io3 + 1
        delta = arr[io2] - arr[io1]
    rcut1 = (arr[io2] + arr[io1]) / float(2.0)
    delta = arr[io3] - arr[io2]
    while delta < rcut_buffer and arr[io3] < max_cut and arr[io2] < max_cut:
        io2 = io2 + 1
        io3 = io3 + 1
        delta = arr[io3] - arr[io2]
    rcut2 = float(arr[io3] + arr[io2]) / float(2.0)
    rcut, rcut_dihed = get_prdf(s=s)
    # rcut_dihed=min(rcut_dihed,max_dihed)
    return rcut, rcut1, rcut2, rcut_dihed


def get_structure_data(s="", c_size=10.0):
    """
    Convert Structure object to a represnetation where only unique sites are keps
    Used for angular distribution terms
    Args:
        s: Structure object
    Returns:
          info with coords,dim,nat,lat,new_symbs
    """
    info = {}
    coords = s.frac_coords
    box = s.lattice.matrix
    dim1 = int(float(c_size) / float(max(abs(box[0])))) + 1
    dim2 = int(float(c_size) / float(max(abs(box[1])))) + 1
    dim3 = int(float(c_size) / float(max(abs(box[2])))) + 1
    dim = [dim1, dim2, dim3]
    dim = np.array(dim)
    all_symbs = [i.symbol for i in s.species]
    nat = len(coords)

    new_nat = nat * dim[0] * dim[1] * dim[2]
    new_coords = np.zeros((new_nat, 3))
    new_symbs = []  # np.chararray((new_nat))

    count = 0
    for i in range(nat):
        for j in range(dim[0]):
            for k in range(dim[1]):
                for l in range(dim[2]):
                    new_coords[count][0] = (coords[i][0] + j) / float(dim[0])
                    new_coords[count][1] = (coords[i][1] + k) / float(dim[1])
                    new_coords[count][2] = (coords[i][2] + l) / float(dim[2])
                    new_symbs.append(all_symbs[i])
                    count = count + 1

    # print ('here4')
    nat = new_nat
    coords = new_coords

    nat = len(coords)  # int(s.composition.num_atoms)
    lat = np.zeros((3, 3))
    lat[0][0] = dim[0] * box[0][0]
    lat[0][1] = dim[0] * box[0][1]
    lat[0][2] = dim[0] * box[0][2]
    lat[1][0] = dim[1] * box[1][0]
    lat[1][1] = dim[1] * box[1][1]
    lat[1][2] = dim[1] * box[1][2]
    lat[2][0] = dim[2] * box[2][0]
    lat[2][1] = dim[2] * box[2][1]
    lat[2][2] = dim[2] * box[2][2]

    info["coords"] = coords
    info["dim"] = dim
    info["nat"] = nat
    info["lat"] = lat
    info["new_symbs"] = new_symbs
    return info


@timeout(timelimit)
def ang_dist1(struct_info={}, c_size=10.0, max_n=500, plot=True, max_cut=5.0, rcut=""):
    """
    Get  angular distribution function upto first neighbor

    Args:
        struct_info: struct information
        max_n: maximum number of neigbors
        c_size: max. cell size
        plot: whether to plot distributions
        max_cut: max. bond cut-off for angular distribution
    Retruns:
         ang_hist1: Angular distribution upto first cut-off
         ang_bins1: angle bins
    """
    coords = struct_info["coords"]
    dim = struct_info["dim"]
    nat = struct_info["nat"]
    lat = struct_info["lat"]
    znm = 0
    bond_arr = []
    deg_arr = []
    nn = np.zeros((nat), dtype="int")
    dist = np.zeros((max_n, nat))
    nn_id = np.zeros((max_n, nat), dtype="int")
    bondx = np.zeros((max_n, nat))
    bondy = np.zeros((max_n, nat))
    bondz = np.zeros((max_n, nat))
    dim05 = [float(1 / 2.0) for i in dim]
    for i in range(nat):
        for j in range(i + 1, nat):
            diff = coords[i] - coords[j]
            for v in range(3):
                if np.fabs(diff[v]) >= dim05[v]:
                    diff[v] = diff[v] - np.sign(diff[v])
            new_diff = np.dot(diff, lat)
            dd = np.linalg.norm(new_diff)
            if dd < rcut and dd >= 0.1:
                nn_index = nn[i]  # index of the neighbor
                nn[i] = nn[i] + 1
                dist[nn_index][i] = dd  # nn_index counter id
                nn_id[nn_index][i] = j  # exact id
                bondx[nn_index][i] = new_diff[0]
                bondy[nn_index][i] = new_diff[1]
                bondz[nn_index][i] = new_diff[2]
                nn_index1 = nn[j]  # index of the neighbor
                nn[j] = nn[j] + 1
                dist[nn_index1][j] = dd  # nn_index counter id
                nn_id[nn_index1][j] = i  # exact id
                bondx[nn_index1][j] = -new_diff[0]
                bondy[nn_index1][j] = -new_diff[1]
                bondz[nn_index1][j] = -new_diff[2]
    # print ('maxnn',max(nn))

    ang_at = {}

    for i in range(nat):
        for in1 in range(nn[i]):
            j1 = nn_id[in1][i]
            for in2 in range(in1 + 1, nn[i]):
                j2 = nn_id[in2][i]
                nm = dist[in1][i] * dist[in2][i]
                if nm != 0:
                    rrx = bondx[in1][i] * bondx[in2][i]
                    rry = bondy[in1][i] * bondy[in2][i]
                    rrz = bondz[in1][i] * bondz[in2][i]
                    cos = float(rrx + rry + rrz) / float(nm)
                    if cos <= -1.0:
                        cos = cos + 0.000001
                    if cos >= 1.0:
                        cos = cos - 0.000001
                    deg = math.degrees(math.acos(cos))
                    ang_at.setdefault(round(deg, 3), []).append(i)
                else:
                    znm = znm + 1
    angs = np.array([float(i) for i in ang_at.keys()])
    norm = np.array([float(len(i)) / float(len(set(i))) for i in ang_at.values()])
    ang_hist1, ang_bins1 = np.histogram(
        angs, weights=norm, bins=np.arange(1, 181.0, 1), density=False
    )

    if plot == True:
        plt.bar(ang_bins1[:-1], ang_hist1)
        plt.savefig("ang1.png")
        plt.close()

    return ang_hist1, ang_bins1


@timeout(timelimit)
def ang_dist2(struct_info={}, c_size=10.0, max_n=500, max_cut=5.0, rcut2="", plot=True):
    """
    Get  angular distribution function upto second neighbor

    Args:
        struct_info: struct information
        max_n: maximum number of neigbors
        c_size: max. cell size
        plot: whether to plot distributions
        max_cut: max. bond cut-off for angular distribution
    Retruns:
         ang_hist2: Angular distribution upto secondt cut-off
         ang_bins2: angle bins
    """
    coords = struct_info["coords"]
    coords = struct_info["coords"]
    dim = struct_info["dim"]
    nat = struct_info["nat"]
    lat = struct_info["lat"]

    ### 2nd neighbors
    znm = 0
    bond_arr = []
    deg_arr = []
    nn = np.zeros((nat), dtype="int")
    dist = np.zeros((max_n, nat))
    nn_id = np.zeros((max_n, nat), dtype="int")
    bondx = np.zeros((max_n, nat))
    bondy = np.zeros((max_n, nat))
    bondz = np.zeros((max_n, nat))
    dim05 = [float(1 / 2.0) for i in dim]
    for i in range(nat):
        for j in range(i + 1, nat):
            diff = coords[i] - coords[j]
            for v in range(3):
                if np.fabs(diff[v]) >= dim05[v]:
                    diff[v] = diff[v] - np.sign(diff[v])
            new_diff = np.dot(diff, lat)
            dd = np.linalg.norm(new_diff)
            if dd < rcut2 and dd >= 0.1:
                nn_index = nn[i]  # index of the neighbor
                nn[i] = nn[i] + 1
                dist[nn_index][i] = dd  # nn_index counter id
                nn_id[nn_index][i] = j  # exact id
                bondx[nn_index, i] = new_diff[0]
                bondy[nn_index, i] = new_diff[1]
                bondz[nn_index, i] = new_diff[2]
                nn_index1 = nn[j]  # index of the neighbor
                nn[j] = nn[j] + 1
                dist[nn_index1][j] = dd  # nn_index counter id
                nn_id[nn_index1][j] = i  # exact id
                bondx[nn_index1, j] = -new_diff[0]
                bondy[nn_index1, j] = -new_diff[1]
                bondz[nn_index1, j] = -new_diff[2]

    # print ('ang at2')
    ang_at = {}

    for i in range(nat):
        for in1 in range(nn[i]):
            j1 = nn_id[in1][i]
            for in2 in range(in1 + 1, nn[i]):
                j2 = nn_id[in2][i]
                nm = dist[in1][i] * dist[in2][i]
                if nm != 0:
                    rrx = bondx[in1][i] * bondx[in2][i]
                    rry = bondy[in1][i] * bondy[in2][i]
                    rrz = bondz[in1][i] * bondz[in2][i]
                    cos = float(rrx + rry + rrz) / float(nm)
                    if cos <= -1.0:
                        cos = cos + 0.000001
                    if cos >= 1.0:
                        cos = cos - 0.000001
                    deg = math.degrees(math.acos(cos))
                    # ang_at.setdefault(deg, []).append(i)
                    ang_at.setdefault(round(deg, 3), []).append(i)
                else:
                    znm = znm + 1
    angs = np.array([float(i) for i in ang_at.keys()])
    norm = np.array([float(len(i)) / float(len(set(i))) for i in ang_at.values()])

    ang_hist2, ang_bins2 = np.histogram(
        angs, weights=norm, bins=np.arange(1, 181.0, 1), density=False
    )
    if plot == True:
        plt.bar(ang_bins2[:-1], ang_hist2)
        plt.savefig("ang2.png")
        plt.close()
    return ang_hist2, ang_bins2


@timeout(timelimit)
def dihedrals(
    struct_info={}, new_symbs=[], max_n=500, c_size=10.0, rcut_dihed="", plot=False
):
    """
    Get dihedral angle distribution
    """

    coords = struct_info["coords"]
    dim = struct_info["dim"]
    nat = struct_info["nat"]
    lat = struct_info["lat"]

    znm = 0
    bond_arr = []
    deg_arr = []
    nn = np.zeros((nat), dtype="int")
    dist = np.zeros((max_n, nat))
    nn_id = np.zeros((max_n, nat), dtype="int")
    bondx = np.zeros((max_n, nat))
    bondy = np.zeros((max_n, nat))
    bondz = np.zeros((max_n, nat))
    dim05 = [float(1 / 2.0) for i in dim]
    for i in range(nat):
        for j in range(i + 1, nat):
            diff = coords[i] - coords[j]
            for v in range(3):
                if np.fabs(diff[v]) >= dim05[v]:
                    diff[v] = diff[v] - np.sign(diff[v])
            new_diff = np.dot(diff, lat)
            dd = np.linalg.norm(new_diff)
            if dd < rcut_dihed and dd >= 0.1:
                # if dd<rcut1 and dd>=0.1:
                nn_index = nn[i]  # index of the neighbor
                nn[i] = nn[i] + 1
                dist[nn_index][i] = dd  # nn_index counter id
                nn_id[nn_index][i] = j  # exact id
                bondx[nn_index, i] = new_diff[0]
                bondy[nn_index, i] = new_diff[1]
                bondz[nn_index, i] = new_diff[2]
                nn_index1 = nn[j]  # index of the neighbor
                nn[j] = nn[j] + 1
                dist[nn_index1][j] = dd  # nn_index counter id
                nn_id[nn_index1][j] = i  # exact id
                bondx[nn_index1, j] = -new_diff[0]
                bondy[nn_index1, j] = -new_diff[1]
                bondz[nn_index1, j] = -new_diff[2]

    dih_at = {}
    for i in range(nat):
        for in1 in range(nn[i]):
            j1 = nn_id[in1][i]
            if j1 > i:
                for in2 in range(nn[i]):  # all other nn of i that are not j
                    j2 = nn_id[in2][i]
                    if j2 != j1:
                        for in3 in range(nn[j1]):  # all other nn of j that are not i
                            j3 = nn_id[in3][j1]
                            if j3 != i:
                                v1 = []
                                v2 = []
                                v3 = []
                                v1.append(bondx[in2][i])
                                v1.append(bondy[in2][i])
                                v1.append(bondz[in2][i])
                                v2.append(-bondx[in1][i])
                                v2.append(-bondy[in1][i])
                                v2.append(-bondz[in1][i])
                                v3.append(-bondx[in3][j1])
                                v3.append(-bondy[in3][j1])
                                v3.append(-bondz[in3][j1])
                                v23 = np.cross(v2, v3)
                                v12 = np.cross(v1, v2)
                                theta = math.degrees(
                                    math.atan2(
                                        np.linalg.norm(v2) * np.dot(v1, v23),
                                        np.dot(v12, v23),
                                    )
                                )
                                if theta < 0.00001:
                                    theta = -theta
                                # print "theta=",theta
                                dih_at.setdefault(round(theta, 3), []).append(i)
    dih = np.array([float(i) for i in dih_at.keys()])
    dih1 = set(dih)
    # print "dih",dih1
    norm = np.array([float(len(i)) / float(len(set(i))) for i in dih_at.values()])

    dih_hist1, dih_bins1 = np.histogram(
        dih, weights=norm, bins=np.arange(1, 181.0, 1), density=False
    )
    if plot == True:
        plt.bar(dih_bins1[:-1], dih_hist1)
        plt.savefig("dihedrals.png")
        plt.close()
    return dih_hist1, dih_bins1


def get_chgdescrp_arr(elm=""):
    """
      Get charge descriptors for an element

      Args:
           elm: element name
      Returns:
             arr: array value
      """
    arr = []

    try:
        f = open(el_chrg_json, "r")
        emdat = json.load(f)
        f.close()
        arr = emdat[elm][0][1]
    except:
        pass
    return arr


def get_descrp_arr_name(elm="Al"):
    """
      Get chemical descriptors for an element

      Args:
           elm: element name
      Returns:
             arr: array value
      """
    arr = []
    try:
        f = open(el_chem_json, "r")
        dat = json.load(f)
        f.close()

        d = dat[elm]
        arr = []
        for k, v in d.items():
            arr.append(k)
    except:
        pass
    return arr


def get_descrp_arr(elm=""):
    """
      Get chemical descriptors for an element

      Args:
           elm: element name
      Returns:
             arr: array value
      """
    arr = []
    try:
        f = open(el_chem_json, "r")
        dat = json.load(f)
        f.close()

        d = dat[elm]
        arr = []
        for k, v in d.items():
            arr.append(v)
        arr = np.array(arr).astype(float)
    except:
        pass
    return arr


def packing_fraction(s=None):
    """
    Get packing fraction

    Args:
         s: Structure object
    Returns:
           packing fraction
    """
    total_rad = 0
    for site in s:
        total_rad += site.specie.atomic_radius ** 3
    pf = np.array([4 * pi * total_rad / (3 * s.volume)])
    return pf[0]


def get_comp_descp(
    struct="",
    jcell=True,
    jmean_chem=True,
    jmean_chg=True,
    jrdf=False,
    jrdf_adf=True,
    print_names=False,
):
    """
        Get chemo-structural CFID decriptors

        Args:
            struct: Structure object
            jcell: whether to use cell-size descriptors
            jmean_chem: whether to use average chemical descriptors
            jmean_chg: whether to use average charge distribution descriptors
            jmean_rdf: whether to use radial distribution descriptors
            jrdf_adf: whether to use radial as well as angle distribution descriptors
            print_names: whether to print names of descriptors
        Returns:
              cat: catenated final descriptors
        """
    cat = []
    s = struct
    # s = get_effective_structure(struct)
    cell = []
    mean_chem = []
    rdf = []
    adf = []
    nn = []
    mean_chg = []
    adfa = []
    adfb = []
    ddf = []

    if jmean_chem == True:
        comp = s.composition
        el_dict = comp.get_el_amt_dict()
        arr = []
        for k, v in el_dict.items():
            des = get_descrp_arr(k)
            arr.append(des)
        mean_chem = np.mean(arr, axis=0)
        # print ('mean_chem',len(mean_chem))

    if jcell == True:
        v_pa = round(float(s.volume) / float(s.composition.num_atoms), 5)
        logv_pa = round(log(float(s.volume) / float(s.composition.num_atoms)), 5)
        pf = round(packing_fraction(s), 5)
        density = round(s.density, 5)
        cell = np.array([v_pa, logv_pa, pf, density])
        # print ('jcell',len(cell))

    if jrdf == True:
        distrdf, bins, bo = get_rdf(s=s)
        rdf = np.array(distrdf)
        # print ('rdf',len(rdf))

    if jrdf_adf == True:
        rcut, rcut1, rcut2, rcut_dihed = get_dist_cutoffs(s=s)
        # print ('rcut,rcut1,rcut2,rcut_dihed',rcut,rcut1,rcut2,rcut_dihed)
        struct_info = get_structure_data(s=s)
        # print ('struct_info',struct_info)
        bins, rdf, nn = get_rdf(s=s)
        try:
            adfa = np.zeros(179)
            adfa, _ = ang_dist1(struct_info=struct_info, plot=False, rcut=rcut_dihed)
        except:
            print(
                "Angular distribution1 part is taking too long a time,setting it to zero"
            )
            pass
        try:
            adfb = np.zeros(179)
            adfb, _ = ang_dist2(struct_info=struct_info, plot=False, rcut2=rcut2)
        except:
            print(
                "Angular distribution2 part is taking too long a time,setting it to zero"
            )
            pass
        try:
            ddf = np.zeros(179)
            ddf, _ = dihedrals(
                struct_info=struct_info, plot=False, rcut_dihed=rcut_dihed
            )
        except:
            print(
                "Dihedral distribution part is taking too long a time,setting it to zero"
            )
            pass
        adfa = np.array(adfa)
        adfb = np.array(adfb)
        rdf = np.array(rdf)
        ddf = np.array(ddf)
        nn = np.array(nn)
        # print ('adfa',len(adfa))
        # print ('ddf',len(ddf))
        # print ('adfb',len(adfb))
        # print ('rdf',len(rdf))
        # print ('nn',len(nn))

    if jmean_chg == True:
        chgarr = []
        for k, v in el_dict.items():
            chg = get_chgdescrp_arr(k)
            chgarr.append(chg)
        mean_chg = np.mean(chgarr, axis=0)
        # print ('mean_chg',len(mean_chg))

    if print_names == True:
        nmes = []
        chem_nm = get_descrp_arr_name()
        for d, nm in zip(
            [mean_chem, cell, mean_chg, rdf, adfa, adfb, ddf, nn],
            ["mean_chem", "cell", "mean_chg", "rdf", "adfa", "adfb", "ddf", "nn"],
        ):
            if d != []:
                for ff, dd in enumerate(d):
                    cat.append(dd)
                    if nm == "mean_chem":
                        tag = chem_nm[ff]
                    else:
                        tag = str(nm) + str("_") + str(ff)
                    nmes.append(str(tag))
        cat = np.array(cat).astype(float)
        # print (nmes,len(nmes))
        return nmes
    else:
        for d, nm in zip(
            [mean_chem, cell, mean_chg, rdf, adfa, adfb, ddf, nn],
            ["mean_chem", "cell", "mean_chg", "rdf", "adfa", "adfb", "ddf", "nn"],
        ):
            if d != []:
                for ff, dd in enumerate(d):
                    cat.append(dd)
        cat = np.array(cat).astype(float)
    ##except:
    # pass
    return cat


def get_chemonly(string=""):
    """
  Get only mean chemical descriptors for a chemical formula, say Al2O3

  Args:
       string: chemical formula, say Al2O3, NaCl etc.
  Returns:
         mean_chem: average chemical descriptors
  """
    comp = Composition(string)
    el_dict = comp.get_el_amt_dict()
    arr = []
    for k, v in el_dict.items():
        des = get_descrp_arr(k)
        arr.append(des)
    mean_chem = np.mean(arr, axis=0)
    return mean_chem


if __name__ == "__main__":

    # chemo-structural features
    s = Structure.from_file("POSCAR")
    ss = s.copy()
    # Making supercell
    ss.make_supercell([2, 3, 4])
    x = get_comp_descp(s)
    xx = get_comp_descp(ss)
    print("len", len(x))
    # print(len(x))
    count = 0
    for i, j in zip(x, xx):
        count = count + 1
        if i != j:
            print(count, i, j)
    # only chemical features for a structure
    y = get_comp_descp(
        struct=s,
        jcell=False,
        jmean_chem=True,
        jmean_chg=False,
        jrdf=False,
        jrdf_adf=False,
        print_names=False,
    )
    print(len(y))

    # chemical features based on composition only
    chm = get_chemonly("Al2O3")
    print(len(chm))

    # charge descriptors
    comp = Composition("Al2O3")
    el_dict = comp.get_el_amt_dict()
    chgarr = []
    for k, v in el_dict.items():
        chg = get_chgdescrp_arr(k)
        chgarr.append(chg)
    mean_chg = np.mean(chgarr, axis=0)
    print(len(mean_chg))



import os, sys, zipfile


def ZipDir(inputDir, outputZip, contents=[]):
    """
    Zip up a directory and preserve symlinks and empty directories
    """
    zipOut = zipfile.ZipFile(outputZip, "w", compression=zipfile.ZIP_DEFLATED)
    tmp = contents
    rootLen = len(os.path.dirname(inputDir))

    def _ArchiveDirectory(parentDirectory):
        # contents = os.listdir(parentDirectory)
        contents = (
            tmp
        )  # ['init.mod', 'potential.mod', 'in.elastic', 'data',  'log.lammps', 'restart.equil','data0' ]
        # store empty directories
        if not contents:
            # http://www.velocityreviews.com/forums/t318840-add-empty-directory-using-zipfile.html
            archiveRoot = parentDirectory[rootLen:].replace("\\", "/").lstrip("/")
            zipInfo = zipfile.ZipInfo(archiveRoot + "/")
            zipOut.writestr(zipInfo, "")
        for item in contents:
            fullPath = os.path.join(parentDirectory, item)
            if os.path.isdir(fullPath) and not os.path.islink(fullPath):
                _ArchiveDirectory(fullPath)
            else:
                archiveRoot = fullPath[rootLen:].replace("\\", "/").lstrip("/")
                if os.path.islink(fullPath):
                    # http://www.mail-archive.com/python-list@python.org/msg34223.html
                    zipInfo = zipfile.ZipInfo(archiveRoot)
                    zipInfo.create_system = 3
                    # long type of hex val of '0xA1ED0000L',
                    # say, symlink attr magic...
                    # zipInfo.external_attr = 0777 << 16L
                    zipInfo.external_attr = "2716663808L"
                    zipOut.writestr(zipInfo, os.readlink(fullPath))
                else:
                    zipOut.write(fullPath, archiveRoot, zipfile.ZIP_DEFLATED)

    _ArchiveDirectory(inputDir)

    zipOut.close()



from __future__ import unicode_literals, print_function

"""
Helper function for running LAMMPS
Used for defects, surface and phonon calculations
"""

from ase import *
import numpy as np
import os

try:
    from phonopy import Phonopy
    from phonopy.file_IO import parse_FORCE_CONSTANTS, write_FORCE_CONSTANTS
    from phonopy.structure.atoms import Atoms as PhonopyAtoms
except:
    pass
import glob
from pymatgen.core.structure import Structure


def get_phonopy_atoms(mat=None):
    """
    Helper function to convert pymatgen structure object to phonopy atoms

    Args:
        mat: pymatgen structure object
    Returns:
           phonopy atoms object
    """
    symbols = [str(site.specie.symbol) for site in mat]
    positions = [site.coords for site in mat]
    cell = mat.lattice.matrix
    p = PhonopyAtoms(symbols=symbols, positions=positions, pbc=True, cell=cell)
    return p


if __name__ == "__main__":
    pos = str(
        os.path.join(
            os.path.dirname(__file__),
            "../vasp/examples/SiOptb88/MAIN-MBJ-bulk@mp_149/POSCAR",
        )
    )
    mat = Structure.from_file(pos)
    p = get_phonopy_atoms(mat)
    print(p)



from __future__ import unicode_literals, print_function
from pymatgen.ext.matproj import MPRester
import json, os, operator
from monty.json import MontyEncoder, MontyDecoder
from pymatgen.core.periodic_table import get_el_sp, Element


def get_struct_from_mp(formula, MAPI_KEY="", all_structs=False):

    if not MAPI_KEY:
        MAPI_KEY = os.environ.get("MAPI_KEY", "")
        if not MAPI_KEY:
            print("API key not provided")
            print(
                "get API KEY from materialsproject and set it to the MAPI_KEY environment variable. aborting ... "
            )
            sys.exit()
    with MPRester(MAPI_KEY) as m:
        data = m.get_data(formula)
        structures = []
        x = {}
        print(
            "\nnumber of structures matching the chemical formula {0} = {1}".format(
                formula, len(data)
            )
        )
        print(
            "The one with the the lowest energy above the hull is returned, unless all_structs is set to True"
        )
        for d in data:
            mpid = str(d["material_id"])
            x[mpid] = d["e_above_hull"]
            if all_structs:
                structure = m.get_structure_by_material_id(mpid)
                structure.sort()
                structures.append(structure)
        if all_structs:
            return structures
        else:
            key = sorted(x.items(), key=operator.itemgetter(1))[0][0]
            print(
                "The id of the material corresponding to the lowest energy above the hull = {0}".format(
                    key
                )
            )
            if key:
                return key, m.get_structure_by_material_id(key)
            else:
                return None


if __name__ == "__main__":
    f = open("chempot_strt.json", "w")
    mem = []
    for i in Element:
        try:
            key, strt = get_struct_from_mp(i.symbol)
            info = {}
            info["element"] = str(i.symbol)
            info["mpid"] = key
            info["structure"] = strt
            mem.append(info)
        except:
            pass
    f.write(json.dumps(mem, cls=MontyEncoder))
    f.close()



from jarvis.db.static.explore_db import get_3d_dataset, get_2d_dataset, get_ml_dataset


def test_j2d():
    data_2d = get_2d_dataset()
    assert (len(data_2d)) == 754


def test_j3d():
    data_3d = get_3d_dataset()
    assert (len(data_3d)) == 32486


def test_jml():
    data_ml = get_ml_dataset()
    assert (len(data_ml)) == 24759



"""
Module to run OptB88vdW based High-throughput calculations
"""
from __future__ import division, unicode_literals, print_function
import os, socket, shutil
from custodian.vasp.jobs import VaspJob
from pymatgen.io.vasp import VaspInput, Vasprun
from pymatgen.io.vasp.outputs import Oszicar
from subprocess import Popen, PIPE
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
import sys, shutil, glob, codecs
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.core.surface import (
    Slab,
    SlabGenerator,
    generate_all_slabs,
    get_symmetrically_distinct_miller_indices,
)
from custodian.vasp.handlers import (
    VaspErrorHandler,
    UnconvergedErrorHandler,
    MeshSymmetryErrorHandler,
    NonConvergingErrorHandler,
    PotimErrorHandler,
)
from pymatgen.symmetry.bandstructure import HighSymmKpath
import json, yaml
from numpy import linalg as LA
import time
from collections import OrderedDict
from jarvis.lammps.jlammps import vac_antisite_def_struct_gen, surfer

# from jarvis.lammps.Surf_Def import vac_antisite_def_struct_gen,surfer
from pymatgen.ext.matproj import MPRester
import subprocess
from pymatgen.core.structure import Structure
from pymatgen.io.vasp.inputs import Incar, Poscar, VaspInput, Potcar, Kpoints

# from pymatgen.io.vasp.sets import MPVaspInputSet, MPNonSCFVaspInputSet
from numpy import matrix
import numpy as np
from pymatgen.io.ase import AseAtomsAdaptor
import operator
from monty.serialization import loadfn, dumpfn
from monty.json import MontyEncoder, MontyDecoder

ncores = 16
try:
    # Define in enenv_variable file
    main_exe = os.environ["vasp_bulk_exe"]
    surf_exe = os.environ["vasp_surf_exe"]
    nw_exe = os.environ["vasp_nw_exe"]
    soc_exe = os.environ["vasp_soc_exe"]
    pot_yaml = str(os.path.join(os.path.dirname(__file__), "Special_POTCAR.yaml"))
    vdw_dat = os.environ["vasp_vdw_dat"]
    json_dat = os.environ["data_json"]
    ncores = int(os.environ["ncores"])
    nnodes = int(os.environ["nnodes"])
    mem = str(os.environ["mem"])
    walltime = str(os.environ["walltime"])
    mp_cmd = str(os.environ["mp_cmd"])
except:
    pass

optb88dict = dict(
    PREC="Accurate",
    ISMEAR=0,
    IBRION=2,
    GGA="BO",
    PARAM1=0.1833333333,
    PARAM2=0.2200000000,
    LUSE_VDW=".TRUE.",
    AGGAC=0.0000,
    EDIFF="1E-7",
    NSW=1,
    NELM=400,
    ISIF=2,
    NPAR=ncores,
    #NPAR=np.sqrt(ncores),
    LCHARG=".FALSE.",
    LWAVE=".FALSE.",
)
functional = "PBE"
use_incar_dict = optb88dict

# Change use_incar_dict based on a functional,
# examples given for LDA and PBE

# functional='LDA'
# ldadict = dict(
#            PREC = 'Accurate',
#            ISMEAR = 0,
#            IBRION=2,
#
#            EDIFF = '1E-7',
#            NSW = 1,
#            NELM = 400,
#            ISIF = 2,
#            NPAR = np.sqrt(ncores),
#            LCHARG = '.FALSE.',
#            LWAVE = '.FALSE.' )
#
# use_incar_dict=ldadict
#
# functional='PBE'
# pbedict = dict(
#            PREC = 'Accurate',
#            ISMEAR = 0,
#            IBRION=2,
#
#            GGA = 'PE',
#
#            EDIFF = '1E-7',
#            NSW = 1,
#            NELM = 400,
#            ISIF = 2,
#            NPAR = np.sqrt(ncores),
#            LCHARG = '.FALSE.',
#            LWAVE = '.FALSE.' )
#
# use_incar_dict=pbedict


def check_polar(file):
    """
    Check if the surface structure is polar
    by comparing atom types at top and bottom

    Args:
         file:Structure object (surface with vacuum)
    Returns:
           polar:True/False   
    """
    up = 0
    dn = 0
    coords = np.array(file.frac_coords)
    z_max = max(coords[:, 2])
    z_min = min(coords[:, 2])
    for site in file:
        if site.frac_coords[2] == z_max:
            up = up + site.specie.number
        if site.frac_coords[2] == z_min:
            dn = dn + site.specie.number
    polar = False
    if up != dn:
        print("polar")
        polar = True
    if up == dn:
        print("Non-polar")
        polar = False
    return polar


def main_outcar(fil=""):
    cnvg = False
    try:
        f = open(fil, "r")
        lines = f.read().splitlines()
        f.close()
        cnvg = False
        for i in lines:
            if "General timing and accounting informations for this job" in i:
                cnvg = True
        # print fil,cnvg
    except:
        pass
    return cnvg


def get_lowest_en_from_mp(formula, MAPI_KEY="", all_structs=False):
    """
    Lowest energy/chemical potential of an element
    from the materialsproject/jarvis database.
    Note: Get the api key from materialsproject/jarvis website.
 
    Args:
        formula: say Al, Ni etc.
        MAPI_KEY: should be defines in the environment
        all_structs: all structures or just stable ones (True/False)
    Returns:
         enp: energy per atom

    """
    if not MAPI_KEY:
        MAPI_KEY = os.environ.get("MAPI_KEY", "")
        if not MAPI_KEY:
            print("API key not provided")
            print(
                "get API KEY from materialsproject and set it to the MAPI_KEY environment variable. aborting ... "
            )
            sys.exit()
    with MPRester(MAPI_KEY) as m:
        data = m.get_data(formula)
        structures = []
        x = {}
        print(
            "\nnumber of structures matching the chemical formula {0} = {1}".format(
                formula, len(data)
            )
        )
        print(
            "The one with the the lowest energy above the hull is returned, unless all_structs is set to True"
        )
        for d in data:
            mpid = str(d["material_id"])
            x[mpid] = d["e_above_hull"]
            if all_structs:
                structure = m.get_structure_by_material_id(mpid)
                structures.append(structure)
        else:
            mineah_key = sorted(x.items(), key=operator.itemgetter(1))[0][0]
            print(
                "The id of the material corresponding to the lowest energy above the hull = {0}".format(
                    mineah_key
                )
            )
            if mineah_key:
                with MPRester(MAPI_KEY) as m:
                    data = m.get_data(mineah_key)
                    x = {}
                    for d in data:
                        x["energy_per_atom"] = str(d["energy_per_atom"])
                        enp = x["energy_per_atom"]
                # return m.get_structure_by_material_id(mineah_key)
                return enp
            else:
                return None


def sum_chem_pot(strt=None):
    """
    Helper function for sump of chemical potential

    Args:
        strt: Structure object
    Returns:
           sum: sum of energy
    """
    sum = 0
    symb = strt.symbol_set
    for el in symb:
        enp = get_lowest_en_from_mp(el)
        sum = float(sum) + float(enp)
    return sum


def run_job(mat=None, incar=None, kpoints=None, jobname="", copy_file=[]):
    """
    Generic function to run a VASP job, error correction implemented using
    custodian package
    A jobname+.json file is produced after successful completion of the job
    A first_cust.py file is generated which is invoked using python command

    Args:
        mat: Poscar object with structure information
        incar: Incar object with control information
        kpoints: Kpoints object with mesh information
        jobname: a uniq name for a job-type, say MAIN-ELAST for elastic properties
        copy_file: copy file from previous runs, say CHGCAR for Non-SCF bandstructure calculations 
    Returns:
        f_energy: final energy
        contcar: path to final relaxed structure
    """
    # hostname=str(socket.gethostname())
    poscar_list = [(mat)]
    cwd = str(os.getcwd())
    job_name = str(mat.comment)
    job_dir = str(jobname)
    run_file = str(os.getcwd()) + str("/") + str(jobname) + str(".json")
    run_dir = str(os.getcwd()) + str("/") + str(jobname)
    if mat.comment.startswith("Surf"):
        [a, b, c] = kpoints.kpts[0]
        kpoints.kpts = [[a, b, 1]]
        try:
            pol = check_polar(mat.structure)
            if pol == True:
                ase_atoms = AseAtomsAdaptor().get_atoms(mat.structure)
                COM = ase_atoms.get_center_of_mass(scaled=True)
                print("COM=", COM)
                print("Found polar surface, will be setting dipole corrections")
                incar.update(
                    {
                        "LDIPOL": ".TRUE.",
                        "IDIPOL": 3,
                        "ISYM": 0,
                        "DIPOL": str(COM[0])
                        + str(" ")
                        + str(COM[2])
                        + str(" ")
                        + str(COM[2]),
                    }
                )
                print("Polar surface encountered in run_job", mat.comment)
        except:
            pass
    wait = False
    json_file = str(jobname) + str(".json")
    print("json should be here=", str(os.getcwd()) + str("/") + str(json_file))
    # print ('json should be=',json_file,run_file,os.getcwd())
    if os.path.exists(str(os.getcwd()) + str("/") + str(json_file)):
        try:
            data_cal = loadfn(
                str(os.getcwd()) + str("/") + str(json_file), cls=MontyDecoder
            )
            tmp_outcar = (
                str(os.getcwd())
                + str("/")
                + str(json_file.split(".json")[0])
                + str("/OUTCAR")
            )
            print("outcar is", tmp_outcar)
            wait = main_outcar(tmp_outcar)  # True
            print("outcar status", wait)
            if wait == True:
                f_energy = data_cal[0]["final_energy"]
                contcar = (
                    str(os.getcwd())
                    + str("/")
                    + str(json_file.split(".json")[0])
                    + str("/CONTCAR")
                )
                return f_energy, contcar
        except:
            pass
    while wait == False:
        print("Setting up POTCAR")
        with open(pot_yaml, "r") as f:
            doc = yaml.load(f)
            pots = doc["POTCAR"]
        new_symb = []
        for el in mat.site_symbols:
            new_symb.append(pots[el])
        # potcar = Potcar(symbols=new_symb,functional=functional)
        try:
            potcar = Potcar(symbols=new_symb, functional=functional)
        except:
            print(
                "JARVIS-ERROR: Could not set POTCAR, check POTCAR yaml file and VASP_PSP_DIR"
            )
            pass
        if not os.path.exists(run_dir):
            print("Starting new job")
            os.makedirs(run_dir)
            os.chdir(run_dir)
            incar.write_file("INCAR")
            potcar.write_file("POTCAR")
            kpoints.write_file("KPOINTS")
            mat.write_file("POSCAR")
            for i in copy_file:
                print("copying", i)
                shutil.copy2(i, "./")
            # f=open('job.out','w')
            cmd = str("python  first_cust.py >out_dat") + "\n"
            cust_file = open("first_cust.py", "w")
            cline = (
                str(
                    "from pymatgen.io.vasp.inputs import Incar, Poscar, VaspInput,Potcar, Kpoints"
                )
                + "\n"
            )
            cust_file.write(cline)
            cline = str("import os,shutil") + "\n"
            cust_file.write(cline)
            cline = str("from custodian.vasp.jobs import VaspJob") + "\n"
            cust_file.write(cline)
            cline = (
                str(
                    "from custodian.vasp.handlers import VaspErrorHandler, UnconvergedErrorHandler,MeshSymmetryErrorHandler, NonConvergingErrorHandler, PotimErrorHandler"
                )
                + "\n"
            )
            cust_file.write(cline)
            cline = (
                str(
                    "from custodian.vasp.validators import VaspFilesValidator,VasprunXMLValidator"
                )
                + "\n"
            )
            cust_file.write(cline)
            cline = str("from custodian.custodian import Custodian") + "\n"
            cust_file.write(cline)
            cline = str('inc=Incar.from_file("INCAR")') + "\n"
            cust_file.write(cline)
            cline = str('pot=Potcar.from_file("POTCAR")') + "\n"
            cust_file.write(cline)
            cline = str('pos=Poscar.from_file("POSCAR")') + "\n"
            cust_file.write(cline)
            cline = str('kp=Kpoints.from_file("KPOINTS")') + "\n"
            cust_file.write(cline)
            cline = str("shutil.copy2('") + vdw_dat + str("','./')") + "\n"
            cust_file.write(cline)
            cline = str('vinput = VaspInput.from_directory(".")') + "\n"
            cust_file.write(cline)
            cline = (
                str("job=VaspJob(['mpirun', '")
                + str(main_exe)
                + str("'], final=False, backup=False)")
                + "\n"
            )
            if mat.comment.startswith("Surf"):
                cline = (
                    str("job=VaspJob(['mpirun',  '")
                    + str(surf_exe)
                    + str("'], final=False, backup=False)")
                    + "\n"
                )
            if "SOC" in jobname:
                cline = (
                    str("job=VaspJob(['mpirun',  '")
                    + str(soc_exe)
                    + str("'], final=False, backup=False)")
                    + "\n"
                )
            cust_file.write(cline)
            cline = (
                str(
                    "handlers = [VaspErrorHandler(), MeshSymmetryErrorHandler(),UnconvergedErrorHandler(), NonConvergingErrorHandler(),PotimErrorHandler()]"
                )
                + "\n"
            )
            cust_file.write(cline)
            cline = str("validators = [VasprunXMLValidator()]") + "\n"
            # cline=str('validators = [VaspFilesValidator()]')+'\n'
            cust_file.write(cline)
            cline = (
                str("c = Custodian(handlers, [job],max_errors=5,validators=validators)")
                + "\n"
            )
            cust_file.write(cline)
            cline = str("c.run()") + "\n"
            cust_file.write(cline)
            cust_file.close()
            print("I AM HERE 2")
            os.system(cmd)
            if os.path.isfile("OUTCAR"):
                try:
                    wait = main_outcar("OUTCAR")  # Vasprun("vasprun.xml").converged
                except:
                    pass
            print("End of the first loop", os.getcwd(), wait)

        else:
            print("Jobs seens to have started before")
            os.chdir(run_dir)
            wait = False
            if os.path.isfile("OUTCAR"):
                try:
                    wait = main_outcar("OUTCAR")  # Vasprun("vasprun.xml").converged
                    # wait=Vasprun("vasprun.xml").converged
                except:
                    pass
            print("Tried to find OUTCAR, wait is=", wait)
            if wait == False:
                incar.write_file("INCAR")
                kpoints.write_file("KPOINTS")
                mat.write_file("POSCAR")
                try:
                    potcar.write_file("POTCAR")
                    print("FOUND OLD CONTCAR in", os.getcwd())
                    old_contcar = Poscar.from_file("CONTCAR")
                    # old_contcar.write_file('POSCAR')
                    copy_cmd = str("cp CONTCAR POSCAR")
                    print("copy_cmd=", copy_cmd)
                    if "ELAST" not in jobname:
                        # Because in ELASTIC calculations structures are deformed
                        os.system(copy_cmd)
                    # time.sleep(3)
                except:
                    pass
                for i in copy_file:
                    print("copying", i)
                    shutil.copy2(i, "./")

                cmd = str("python  first_cust.py >out_dat") + "\n"
                cust_file = open("first_cust.py", "w")
                cline = (
                    str(
                        "from pymatgen.io.vasp.inputs import Incar, Poscar, VaspInput,Potcar, Kpoints"
                    )
                    + "\n"
                )
                cust_file.write(cline)
                cline = str("import os,shutil") + "\n"
                cust_file.write(cline)
                cline = str("from custodian.vasp.jobs import VaspJob") + "\n"
                cust_file.write(cline)
                cline = (
                    str(
                        "from custodian.vasp.handlers import VaspErrorHandler, UnconvergedErrorHandler,MeshSymmetryErrorHandler, NonConvergingErrorHandler, PotimErrorHandler"
                    )
                    + "\n"
                )
                cust_file.write(cline)
                cline = (
                    str(
                        "from custodian.vasp.validators import VaspFilesValidator,VasprunXMLValidator"
                    )
                    + "\n"
                )
                cust_file.write(cline)
                cline = str("from custodian.custodian import Custodian") + "\n"
                cust_file.write(cline)
                cline = str('inc=Incar.from_file("INCAR")') + "\n"
                cust_file.write(cline)
                cline = str('pot=Potcar.from_file("POTCAR")') + "\n"
                cust_file.write(cline)
                cline = str('pos=Poscar.from_file("POSCAR")') + "\n"
                cust_file.write(cline)
                cline = str('kp=Kpoints.from_file("KPOINTS")') + "\n"
                cust_file.write(cline)
                cline = str("shutil.copy2('") + vdw_dat + str("','./')") + "\n"
                cust_file.write(cline)
                cline = str('vinput = VaspInput.from_directory(".")') + "\n"
                cust_file.write(cline)
                cline = (
                    str("job=VaspJob(['mpirun', '")
                    + str(main_exe)
                    + str("'], final=False, backup=False)")
                    + "\n"
                )
                if mat.comment.startswith("Surf"):
                    cline = (
                        str("job=VaspJob(['mpirun',  '")
                        + str(surf_exe)
                        + str("'], final=False, backup=False)")
                        + "\n"
                    )
                if "SOC" in jobname:
                    cline = (
                        str("job=VaspJob(['mpirun',  '")
                        + str(soc_exe)
                        + str("'], final=False, backup=False)")
                        + "\n"
                    )
                cust_file.write(cline)
                cline = (
                    str(
                        "handlers = [VaspErrorHandler(), MeshSymmetryErrorHandler(),UnconvergedErrorHandler(), NonConvergingErrorHandler(),PotimErrorHandler()]"
                    )
                    + "\n"
                )
                cust_file.write(cline)
                # cline=str('validators = [VaspFilesValidator()]')+'\n'
                cline = str("validators = [VasprunXMLValidator()]") + "\n"
                cust_file.write(cline)
                cline = (
                    str(
                        "c = Custodian(handlers, [job],max_errors=5,validators=validators)"
                    )
                    + "\n"
                )
                cust_file.write(cline)
                cline = str("c.run()") + "\n"
                cust_file.write(cline)
                cust_file.close()
                os.system(cmd)
                if os.path.isfile("OUTCAR"):
                    try:
                        wait = main_outcar("OUTCAR")  # Vasprun("vasprun.xml").converged
                        # wait=Vasprun("vasprun.xml").converged
                    except:
                        pass
    f_energy = "na"
    enp = "na"
    contcar = str(os.getcwd()) + str("/") + str("CONTCAR")
    final_str = Structure.from_file(contcar)
    try:
        oszicar = Oszicar("OSZICAR")
        f_energy = float(oszicar.final_energy)
        enp = float(oszicar.final_energy) / float(final_str.composition.num_atoms)
    except:
        print("Error in OSZICAR file during re-run jpb")
        pass
    natoms = final_str.composition.num_atoms
    os.chdir("../")
    if wait == True:
        data_cal = []
        data_cal.append(
            {
                "jobname": jobname,
                "poscar_initial": mat.as_dict(),
                "poscar_final": final_str.as_dict(),
                "incar": incar.as_dict(),
                "kpoints": kpoints.as_dict(),
                "final_energy": (f_energy),
                "contcar": final_str.as_dict(),
            }
        )
        json_file = str(jobname) + str(".json")
        f_json = open(json_file, "w")
        f_json.write(json.dumps(data_cal, indent=4, cls=MontyEncoder))
        f_json.close()
        print("Wrote json file", contcar)
        return f_energy, contcar


def Auto_Kpoints(mat=None, length=20):
    """
    Geting Kpoints object from structure and line-density

    Args:
         mat: Poscar object with structure information
         length: line-density
    Returns:
         kpp: Kpoint object
    """

    b1 = LA.norm(
        np.array(mat.structure.lattice.reciprocal_lattice_crystallographic.matrix[0])
    )
    b2 = LA.norm(
        np.array(mat.structure.lattice.reciprocal_lattice_crystallographic.matrix[1])
    )
    b3 = LA.norm(
        np.array(mat.structure.lattice.reciprocal_lattice_crystallographic.matrix[2])
    )

    n1 = int(max(1, length * b1 + 0.5))
    n2 = int(max(1, length * b2 + 0.5))
    n3 = int(max(1, length * b3 + 0.5))
    kpp = Kpoints.gamma_automatic(kpts=(n1, n2, n3))
    return kpp


def converg_encut(encut=500, mat=None):
    """
    Function to converg plane-wave cut-off

    Args:
        encut: intial cutoff
        mat: Poscar object
    Returns:
           encut: converged cut-off
    """
    en1 = -10000
    encut1 = encut
    convg_encut1 = False
    convg_encut2 = False

    while convg_encut2 != True:
        # while convg_encut1 !=True and  convg_encut2 !=True:
        tol = 0.001  # change 0.001
        encut_list = []
        encut_list.append(encut)
        length = 10
        encut1 = encut + 50
        incar_dict = use_incar_dict
        incar_dict.update({"ENCUT": encut})
        # print (use_incar_dict)
        incar = Incar.from_dict(incar_dict)
        kpoints = Auto_Kpoints(mat=mat, length=length)
        print(
            "running smart_converge for",
            str(mat.comment) + str("-") + str("ENCUT") + str("-") + str(encut),
        )
        en2, contc = run_job(
            mat=mat,
            incar=incar,
            kpoints=kpoints,
            jobname=str("ENCUT") + str(mat.comment) + str("-") + str(encut),
        )
        while abs(en2 - en1) > tol:
            en1 = en2
            encut1 = encut + 50
            encut_list.append(encut)
            print("Incrementing encut", encut)
            incar_dict = use_incar_dict
            incar_dict.update({"ENCUT": encut1})
            incar = Incar.from_dict(incar_dict)
            print(
                "running smart_converge for",
                str(mat.comment) + str("-") + str("ENCUT") + str("-") + str(encut),
            )
            en2, contc = run_job(
                mat=mat,
                incar=incar,
                kpoints=kpoints,
                jobname=str("ENCUT") + str(mat.comment) + str("-") + str(encut),
            )
        convg_encut1 = True

        # Some extra points to check
        print("Some extra points to check for ENCUT")

        encut2 = encut1 + 50
        incar["ENCUT"] = encut2
        en3, contc = run_job(
            mat=mat,
            incar=incar,
            kpoints=kpoints,
            jobname=str("ENCUT") + str(mat.comment) + str("-") + str(encut2),
        )

        encut3 = encut2 + 50
        incar["ENCUT"] = encut3
        en4, contc = run_job(
            mat=mat,
            incar=incar,
            kpoints=kpoints,
            jobname=str("ENCUT") + str(mat.comment) + str("-") + str(encut3),
        )

        encut4 = encut3 + 50
        incar["ENCUT"] = encut4
        en5, contc = run_job(
            mat=mat,
            incar=incar,
            kpoints=kpoints,
            jobname=str("ENCUT") + str(mat.comment) + str("-") + str(encut4),
        )

        encut5 = encut4 + 50
        incar["ENCUT"] = encut5
        en6, contc = run_job(
            mat=mat,
            incar=incar,
            kpoints=kpoints,
            jobname=str("ENCUT") + str(mat.comment) + str("-") + str(encut5),
        )

        encut6 = encut5 + 50
        incar["ENCUT"] = encut6
        en7, contc = run_job(
            mat=mat,
            incar=incar,
            kpoints=kpoints,
            jobname=str("ENCUT") + str(mat.comment) + str("-") + str(encut6),
        )

        # if en3-en2>tol or en4-en2>tol or en5-en2>tol or en6-en2>tol or en7-en2>tol:
        # if abs(en3-en2)>tol and abs(en4-en2)>tol and abs(en5-en2)>tol and abs(en6-en2)>tol and abs(en7-en2)>tol:
        if (
            abs(en3 - en2) > tol
            or abs(en4 - en2) > tol
            or abs(en5 - en2) > tol
            or abs(en6 - en2) > tol
            or abs(en7 - en2) > tol
        ):

            en1 = en3
            encut = encut1
            fen = open("EXTRA_ENCUT", "w")
            line = str("Extra ENCUT needed ") + str(encut) + "\n"
            fen.write(line)
            fen.close()
        else:
            print("ENCUT convergence achieved for ", mat.comment, encut)
            convg_encut2 = True
    return encut


def converg_kpoints(length=0, mat=None):
    """
    Function to converg K-points

    Args:
        lenght: K-point line density
        mat: Poscar object with structure information
    Returns:
           length1: K-point line density
    """

    en1 = -10000
    encut = 550
    convg_kp1 = False
    convg_kp2 = False
    length1 = length
    kp_list = []
    while convg_kp2 != True:
        # while convg_kp1 !=True and  convg_kp2 !=True:
        tol = 0.001  # change 0.001
        incar_dict = use_incar_dict
        incar_dict.update({"ENCUT": encut})
        incar = Incar.from_dict(incar_dict)
        length1 = length1 + 5
        print("Incrementing length", length1)
        kpoints = Auto_Kpoints(mat=mat, length=length1)
        mesh = kpoints.kpts[0]
        if mesh not in kp_list:
            kp_list.append(mesh)
            en2, contc = run_job(
                mat=mat,
                incar=incar,
                kpoints=kpoints,
                jobname=str("KPOINTS") + str(mat.comment) + str("-") + str(length1),
            )

            while abs(en2 - en1) > tol:
                en1 = en2
                print("Incrementing length", length1)
                incar_dict = use_incar_dict
                incar_dict.update({"ENCUT": encut})
                incar = Incar.from_dict(incar_dict)
                while mesh in kp_list:
                    length1 = length1 + 5
                    ##Assuming you are not super unlucky
                    kpoints = Auto_Kpoints(mat=mat, length=length1)
                    mesh = kpoints.kpts[0]

                kpoints = Auto_Kpoints(mat=mat, length=length1)
                mesh = kpoints.kpts[0]
                if mesh not in kp_list:
                    kp_list.append(mesh)
                    en2, contc = run_job(
                        mat=mat,
                        incar=incar,
                        kpoints=kpoints,
                        jobname=str("KPOINTS")
                        + str(mat.comment)
                        + str("-")
                        + str(length1),
                    )
                else:
                    length1 = length1 + 5
                    ##Assuming you are not super unlucky
                    kpoints = Auto_Kpoints(mat=mat, length=length1)
                    mesh = kpoints.kpts[0]
                    kp_list.append(mesh)
                    en2, contc = run_job(
                        mat=mat,
                        incar=incar,
                        kpoints=kpoints,
                        jobname=str("KPOINTS")
                        + str(mat.comment)
                        + str("-")
                        + str(length1),
                    )
            convg_kp1 = True

            # Some extra points to check
            print("Some extra points to check for KPOINTS")
            length3 = length1 + 5
            kpoints = Auto_Kpoints(mat=mat, length=length3)
            mesh = kpoints.kpts[0]
            kp_list.append(mesh)
            en3, contc = run_job(
                mat=mat,
                incar=incar,
                kpoints=kpoints,
                jobname=str("KPOINTS") + str(mat.comment) + str("-") + str(length3),
            )

            length4 = length3 + 5
            kpoints = Auto_Kpoints(mat=mat, length=length4)
            mesh = kpoints.kpts[0]
            kp_list.append(mesh)
            en4, contc = run_job(
                mat=mat,
                incar=incar,
                kpoints=kpoints,
                jobname=str("KPOINTS") + str(mat.comment) + str("-") + str(length4),
            )

            length5 = length4 + 5
            kpoints = Auto_Kpoints(mat=mat, length=length5)
            mesh = kpoints.kpts[0]
            kp_list.append(mesh)
            en5, contc = run_job(
                mat=mat,
                incar=incar,
                kpoints=kpoints,
                jobname=str("KPOINTS") + str(mat.comment) + str("-") + str(length5),
            )

            length6 = length5 + 5
            kpoints = Auto_Kpoints(mat=mat, length=length6)
            mesh = kpoints.kpts[0]
            kp_list.append(mesh)
            en6, contc = run_job(
                mat=mat,
                incar=incar,
                kpoints=kpoints,
                jobname=str("KPOINTS") + str(mat.comment) + str("-") + str(length6),
            )
            length7 = length6 + 5
            kpoints = Auto_Kpoints(mat=mat, length=length7)
            mesh = kpoints.kpts[0]
            kp_list.append(mesh)
            en7, contc = run_job(
                mat=mat,
                incar=incar,
                kpoints=kpoints,
                jobname=str("KPOINTS") + str(mat.comment) + str("-") + str(length7),
            )

            # if en3-en2>tol or en4-en2>tol or en5-en2>tol or en6-en2>tol or en7-en2>tol:
            # if abs(en3-en2)>tol and abs(en4-en2)>tol and abs(en5-en2)>tol and abs(en6-en2)>tol and abs(en7-en2)>tol:
            if (
                abs(en3 - en2) > tol
                or abs(en4 - en2) > tol
                or abs(en5 - en2) > tol
                or abs(en6 - en2) > tol
                or abs(en7 - en2) > tol
            ):
                fkp = open("EXTRA_KPOINTS", "w")
                line = str("Extra KPOINTS needed ") + str(length1) + "\n"
                fkp.write(line)
                line = (
                    str("en2 length1 ")
                    + str(" ")
                    + str(en2)
                    + str(" ")
                    + str(length1)
                    + "\n"
                )
                fkp.write(line)
                line = (
                    str("en3 length3 ")
                    + str(" ")
                    + str(en3)
                    + str(" ")
                    + str(length3)
                    + "\n"
                )
                fkp.write(line)
                line = (
                    str("en4 length4 ")
                    + str(" ")
                    + str(en4)
                    + str(" ")
                    + str(length4)
                    + "\n"
                )
                fkp.write(line)
                line = (
                    str("en5 length5 ")
                    + str(" ")
                    + str(en5)
                    + str(" ")
                    + str(length5)
                    + "\n"
                )
                fkp.write(line)
                fkp.close()
                en1 = en3
                length1 = length3
            else:
                print("KPOINTS convergence achieved for ", mat.comment, length1)
                convg_kp2 = True

    return length1


def smart_converge(
    mat=None,
    encut="",
    leng="",
    band_str=True,
    elast_prop=True,
    optical_prop=True,
    mbj_prop=True,
    spin_orb=False,
    phonon=False,
    surf_en=False,
    def_en=False,
):
    """
    Main function to converge k-points/cut-off
    optimize structure, and run subsequent property calculations

    Args:
         mat: Poscar object with structure information
         encut: if '' then automataic convergence, else use defined fixed-cutoff
         leng: if '' then automataic convergence, else use defined fixed-line density
         band_str: if True then do band-structure calculations along high-symmetry points
         elast_prop: if True then do elastic property calculations using finite-difference
         optical_prop: if True do frequency dependent dielectric function calculations using he independent-particle (IP) approximation
         mbj_prop: if True do METAGGA-TBmBJ optical property calculations with IP approximation
         spin_orb: if True do spin-orbit calculations
         phonon: if True do phonon calculations using DFPT
         surf_en: if True do surface enrgy calculations 
         def_en: if True do defect enrgy calculations 
    Returns:
         en2: final energy
         mat_f: final structure
    """
    if encut == "":
        encut = converg_encut(encut=500, mat=mat)

    if leng == "":
        leng = converg_kpoints(length=0, mat=mat)

    kpoints = Auto_Kpoints(mat=mat, length=leng)
    isif = 2
    commen = str(mat.comment)
    lcharg = ".FALSE."
    if commen.split("@")[0] == "bulk":
        isif = 3
        lcharg = ".TRUE."
    if commen.split("@")[0] == "sbulk":
        isif = 3
    incar_dict = use_incar_dict
    incar_dict.update(
        {
            "ENCUT": encut,
            "EDIFFG": -1e-3,
            "ISIF": 3,
            "NEDOS": 5000,
            "NSW": 500,
            "NELM": 500,
            "LORBIT": 11,
            "LVTOT": ".TRUE.",
            "LVHAR": ".TRUE.",
            "ISPIN": 2,
            "LCHARG": ".TRUE.",
        }
    )
    incar = Incar.from_dict(incar_dict)
    try:
        if mat.comment.startswith("Mol"):
            incar.update({"ISIF": "2"})
    except:
        print("Mol error")
        pass
    # if commen.startswith('Surf-') :
    #   pol=check_polar(mat_f.structure)
    #   if pol==True:
    #        ase_atoms = AseAtomsAdaptor().get_atoms(mat_f.structure)
    #        COM=ase_atoms.get_center_of_mass(scaled=True)
    #        incar.update({"LDIPOL": '.TRUE.',"IDIPOL":4,"ISYM": 0,"DIPOL":COM})
    print("running smart_converge for", str(mat.comment) + str("-") + str("MAIN-RELAX"))
    cwd = str(os.getcwd())
    en2, contc = run_job(
        mat=mat,
        incar=incar,
        kpoints=kpoints,
        jobname=str("MAIN-RELAX") + str("-") + str(mat.comment),
    )
    os.chdir(cwd)
    path = str(contc.split("/CONTCAR")[0]) + str("/vasprun.xml")
    v = open(path, "r").readlines()
    for line in v:
        if "NBANDS" in line:
            nbands = int(line.split(">")[1].split("<")[0])
            print("nbands=", nbands)
            break
    strt = Structure.from_file(contc)
    mat_f = Poscar(strt)
    mat_f.comment = str(mat.comment)
    if band_str == True:
        incar_dict = use_incar_dict
        incar_dict.update(
            {
                "ISPIN": 2,
                "NEDOS": 5000,
                "LORBIT": 11,
                "IBRION": 1,
                "ENCUT": encut,
                "NBANDS": int(nbands) + 10,
            }
        )
        incar = Incar.from_dict(incar_dict)
        kpath = HighSymmKpath(mat_f.structure)
        frac_k_points, k_points_labels = kpath.get_kpoints(
            line_density=20, coords_are_cartesian=False
        )
        kpoints = Kpoints(
            comment="Non SCF run along symmetry lines",
            style=Kpoints.supported_modes.Reciprocal,
            num_kpts=len(frac_k_points),
            kpts=frac_k_points,
            labels=k_points_labels,
            kpts_weights=[1] * len(frac_k_points),
        )

        try:
            print("running MAIN-BAND")
            kpoints = mpvis.get_kpoints(mat_f.structure)
            en2B, contcB = run_job(
                mat=mat_f,
                incar=incar,
                kpoints=kpoints,
                jobname=str("MAIN-BAND") + str("-") + str(mat_f.comment),
            )
        # kpoints=mpvis.get_kpoints(mat_f.structure)
        # en2B,contcB=run_job(mat=mat_f,incar=incar,kpoints=kpoints,jobname=str('MAIN-BAND')+str('-')+str(mat_f.comment))
        except:
            print("No band str calc.")
            if str(os.getcwd) != cwd:
                print("Changing directory")
                line = str("cd ") + str(cwd)
                os.chdir(cwd)
                print(os.getcwd())

            pass
    os.chdir(cwd)
    if surf_en == True:
        incar_dict = use_incar_dict
        incar_dict.update(
            {"ENCUT": encut, "NEDOS": 5000, "IBRION": 1, "NSW": 500, "LORBIT": 11}
        )
        incar = Incar.from_dict(incar_dict)
        surf = surfer(mat=mat_f.structure, layers=3)
        for i in surf:

            try:
                print("running MAIN-BAND")
                # NSCF
                # chg_file=str(contc).replace('CONTCAR','CHGCAR')
                # print ('chrfile',chg_file)
                # shutil.copy2(chg_file,'./')
                kpoints = Auto_Kpoints(mat=i, length=leng)
                en2s, contcs = run_job(
                    mat=i,
                    incar=incar,
                    kpoints=kpoints,
                    jobname=str("Surf_en-")
                    + str(i.comment)
                    + str("-")
                    + str(mat_f.comment),
                )
                # kpoints=mpvis.get_kpoints(mat_f.structure)
                # en2B,contcB=run_job(mat=mat_f,incar=incar,kpoints=kpoints,jobname=str('MAIN-BAND')+str('-')+str(mat_f.comment))
            except:
                pass
    os.chdir(cwd)
    if def_en == True:
        incar_dict = use_incar_dict
        incar_dict.update(
            {"ENCUT": encut, "NEDOS": 5000, "IBRION": 1, "NSW": 500, "LORBIT": 11}
        )
        incar = Incar.from_dict(incar_dict)
        # surf=surfer(mat=mat_f.structure,layers=3)
        vac = vac_antisite_def_struct_gen(cellmax=3, struct=mat_f.structure)
        for i in vac:
            try:
                print("running MAIN-vac")
                kpoints = Auto_Kpoints(mat=i, length=leng)
                en2d, contcd = run_job(
                    mat=i,
                    incar=incar,
                    kpoints=kpoints,
                    jobname=str("Def_en-")
                    + str(i.comment)
                    + str("-")
                    + str(mat_f.comment),
                )
            # kpoints=mpvis.get_kpoints(mat_f.structure)
            # en2B,contcB=run_job(mat=mat_f,incar=incar,kpoints=kpoints,jobname=str('MAIN-BAND')+str('-')+str(mat_f.comment))
            except:
                pass
    os.chdir(cwd)
    # surf=surfer(mat=strt,layers=layers)
    # surf=surfer(mat=strt,layers=layers)
    if spin_orb == True:
        # chg_file=str(contc).replace('CONTCAR','CHGCAR')
        # print ('chrfile',chg_file)
        # shutil.copy2(chg_file,'./')
        incar_dict = use_incar_dict
        incar_dict.update(
            {
                "ENCUT": encut,
                "NPAR": ncores,
                "GGA_COMPAT": ".FALSE.",
                "LSORBIT": ".TRUE.",
                "IBRION": 1,
                "ISYM": 0,
                "NEDOS": 5000,
                "IBRION": 1,
                "NSW": 500,
                "LORBIT": 11,
            }
        )
        incar = Incar.from_dict(incar_dict)
        sg_mat = SpacegroupAnalyzer(mat_f.structure)
        mat_cvn = sg_mat.get_conventional_standard_structure()
        mat_cvn.sort()
        kpoints = Auto_Kpoints(mat=Poscar(mat_cvn), length=leng / 2)
        try:
            en2S, contcS = run_job(
                mat=Poscar(mat_cvn),
                incar=incar,
                kpoints=kpoints,
                jobname=str("MAIN-SOC") + str("-") + str(mat_f.comment),
            )
        except:
            pass
    os.chdir(cwd)
    if optical_prop == True:
        incar_dict = use_incar_dict
        incar_dict.update(
            {
                "NEDOS": 5000,
                "LORBIT": 11,
                "IBRION": 1,
                "ENCUT": encut,
                "NBANDS": 3 * int(nbands),
                "LOPTICS": ".TRUE.",
            }
        )
        incar = Incar.from_dict(incar_dict)
        kpoints = Auto_Kpoints(mat=mat_f, length=leng)
        try:
            en2OP, contcOP = run_job(
                mat=mat_f,
                incar=incar,
                kpoints=kpoints,
                jobname=str("MAIN-OPTICS") + str("-") + str(mat_f.comment),
            )
        except:
            pass
    os.chdir(cwd)
    if mbj_prop == True:
        incar_dict = use_incar_dict
        incar_dict.update(
            {
                "NEDOS": 5000,
                "LORBIT": 11,
                "IBRION": 1,
                "ENCUT": encut,
                "NBANDS": 3 * int(nbands),
                "LOPTICS": ".TRUE.",
                "METAGGA": "MBJ",
                "ISYM": 0,
                "SIGMA": 0.1,
            }
        )
        incar = Incar.from_dict(incar_dict)
        kpoints = Auto_Kpoints(mat=mat_f, length=leng)
        try:
            en2OP, contcOP = run_job(
                mat=mat_f,
                incar=incar,
                kpoints=kpoints,
                jobname=str("MAIN-MBJ") + str("-") + str(mat_f.comment),
            )
        except:
            pass
    os.chdir(cwd)

    if elast_prop == True:
        incar_dict = use_incar_dict
        incar_dict.update(
            {
                "NEDOS": 5000,
                "IBRION": 6,
                "ENCUT": 1.3 * float(encut),
                "ISIF": 3,
                "POTIM": 0.015,
                "NPAR": ncores,
                "ISPIN": 2,
            }
        )
        incar = Incar.from_dict(incar_dict)
        sg_mat = SpacegroupAnalyzer(mat_f.structure)
        mat_cvn = sg_mat.get_conventional_standard_structure()
        mat_cvn.sort()
        kpoints = Auto_Kpoints(mat=Poscar(mat_cvn), length=leng)
        try:
            en2E, contcE = run_job(
                mat=Poscar(mat_cvn),
                incar=incar,
                kpoints=kpoints,
                jobname=str("MAIN-ELASTIC") + str("-") + str(mat_f.comment),
            )
        except:
            pass
    os.chdir(cwd)

    if phonon == True:
        incar_dict = use_incar_dict
        incar_dict.update(
            {
                "IBRION": 8,
                "ENCUT": float(encut),
                "ISYM": 0,
                "ADDGRID": ".TRUE.",
                "EDIFF": 1e-09,
                "LORBIT": 11,
            }
        )
        incar = Incar.from_dict(incar_dict)
        kpoints = Auto_Kpoints(mat=mat_f, length=leng)
        mat_pho = make_big(poscar=mat_f, size=11.0)
        try:
            en2P, contcP = run_job(
                mat=mat_pho,
                incar=incar,
                kpoints=kpoints,
                jobname=str("MAIN-PHO8") + str("-") + str(mat_f.comment),
            )
        except:
            pass
    os.chdir(cwd)

    # if Raman_calc==True:
    #    Raman(strt=mat_f,encut=encut,length=leng)
    os.chdir(cwd)
    return en2, mat_f


def smart_vac(strt=None, tol=0.1):
    """
    Umbrell function for vacancy formation energies with convergence

    Args:
       strt: Structure object
       tol: defect energy convergence tolerance in eV
    Returns:
          def_list: list of defect energies
          def_header_list: list of defect names
           
    """
    vac_arr = []
    sg_mat = SpacegroupAnalyzer(strt)
    mat_cvn = sg_mat.get_conventional_standard_structure()
    mat_cvn.sort()

    cellmax = 1  # int(mat_cvn.composition.num_atoms)+int(mat_cvn.ntypesp)#5
    ase_atoms = AseAtomsAdaptor().get_atoms(mat_cvn)
    ase_atoms = ase_atoms * (cellmax, cellmax, cellmax)
    # if len(ase_atoms) >200:
    #   cellmax=1
    # else:
    #   cellmax=2
    # cellmax=int(mat_cvn.composition.num_atoms)+int(mat_cvn.ntypesp)#5
    print("type of trt is= celmmax", type(strt), cellmax)
    try:
        print("int(strt.composition.num_atoms)", int(strt.composition.num_atoms))
        print(int(strt.ntypesp))
    except:
        pass
    # cellmax=int(strt.composition.num_atoms)+int(strt.ntypesp)
    vac_done = 0
    vac = vac_antisite_def_struct_gen(cellmax=cellmax, struct=strt)
    def_list = [100000 for y in range(len(vac) - 1)]
    while vac_done != 1:
        vac = vac_antisite_def_struct_gen(cellmax=cellmax, struct=strt)
        if vac not in vac_arr:
            vac_arr.append(vac)
            print(
                "in smart_vac(strt=None), cellmax,vac,vac_done=", cellmax, vac, vac_done
            )
            def_list2, header_list = def_energy(vac=vac)
            diff = matrix(def_list) - matrix(def_list2)
            diff_arr = np.array(diff).flatten()
            print("in smart_vac(strt=None diff_arr=", diff_arr)
            if any(diff_arr) > tol:
                # for el in diff_arr:
                #     if abs(el)>tol :
                # print ("in smart_vac(strt=None abs_el=",abs(el))
                vac_done = 0
                cellmax = cellmax + 1
                ase_atoms = AseAtomsAdaptor().get_atoms(mat_cvn)
                ase_atoms = ase_atoms * (cellmax, cellmax, cellmax)
                if len(ase_atoms) > 100:
                    vac_done = 1
                def_list = def_list2
            else:
                vac_done = 1
    #        cellmax=cellmax+1
    return def_list, header_list


def def_energy(vac=[]):
    """
    Calculation of vacancy formation energies

    Args:
        vac: list of Poscar vacancy structure objects
    Returns:
          def_list: list of defect energies
          def_header_list: list of defect names
    """

    def_list = []
    header_list = []
    fi = str("DEF.INFO")
    f = open(fi, "w")
    for v in vac:
        enp, contc = smart_converge(mat=v)
        # enp,contc=run_job(mat=v,incar=incar,kpoints=kpoints,jobname=str(v.comment))
        strt = Structure.from_file(contc)
        comm = str(v.comment)
        print("running def_energy for =", comm)
        header_list.append(comm)
        if comm.split("@")[0] == "bulk":
            gs_energy = float(enp) * float(strt.composition.num_atoms)
            print("in def_energy gs_energy for", comm, gs_energy)
        else:
            chem_pot = sum_chem_pot(strt)
            if comm.startswith("intl"):
                chem_pot = 0.0 - float(chem_pot)
            def_en = (
                (float(enp) * float(strt.composition.num_atoms))
                - float(gs_energy)
                + chem_pot
            )
            print("in def_energy def_en for", comm, def_en, "chem_pot", chem_pot)
            def_list.append(def_en)
            print(v.comment, "=", def_en)
            line = str(v.comment) + str("=") + str(def_en) + "\n"
            f.write(line)
    f.close()
    return def_list, header_list


def smart_surf(strt=None, tol=0.1):
    """
    Umbrell function for surface energies with convergence

    Args:
       strt: Structure object
       tol: surface energy convergence tolerance in eV
    Returns:
          surf_list: list of surface energies
          surf_header_list: list of surface names
           
    """
    sg_mat = SpacegroupAnalyzer(strt)
    mat_cvn = sg_mat.get_conventional_standard_structure()
    mat_cvn.sort()
    layers = 2
    indices = get_symmetrically_distinct_miller_indices(mat_cvn, 1)
    ase_atoms = AseAtomsAdaptor().get_atoms(mat_cvn)
    for i in indices:
        ase_slab = surface(ase_atoms, i, layers)
        ase_slab.center(vacuum=15, axis=2)
        if len(ase_slab) < 50:
            layers = 3
    surf_arr = []
    surf_done = 0
    surf = surfer(mat=strt, layers=layers)
    surf_list = [100000 for y in range(len(surf) - 1)]
    print("in smart_surf :surf,surf_list=", surf, surf_list)
    while surf_done != 1:
        layers = layers + 1
        indices = get_symmetrically_distinct_miller_indices(mat_cvn, 1)
        ase_atoms = AseAtomsAdaptor().get_atoms(mat_cvn)
        for i in indices:
            ase_slab = surface(ase_atoms, i, layers)
            ase_slab.center(vacuum=15, axis=2)
            if len(ase_slab) > 100:
                surf_done = 1
            if (ase_slab.get_cell()[2][2]) > 40:
                surf_done = 1
        surf = surfer(mat=strt, layers=layers)
        if surf not in surf_arr:
            surf_arr.append(surf)
            surf_list2, surf_header_list = surf_energy(surf=surf)
            print("in smart_surf :surf2,surf_list2=", surf_list2, surf_header_list)
            diff = matrix(surf_list) - matrix(surf_list2)
            print(
                "in smart_surf :surf3,surf_list3=",
                matrix(surf_list),
                matrix(surf_list2),
            )
            diff_arr = np.array(diff).flatten()
            if any(diff_arr) > tol:
                # for el in diff_arr:
                #    if abs(el)>tol :
                #        print ("in smart_surf :abs el=",abs(el))
                surf_done = 0
                surf_list = surf_list2
            else:
                surf_done = 1
    return surf_list, surf_header_list


def surf_energy(surf=[]):
    """
    Helper function for surface energies

    Args:
        surf: list of Poscar surface objects
    Returns:
          surf_list: list of surface energies
          surf_header_list: list of surface names
          
    """

    surf_list = []
    surf_header_list = []
    fi = str("SURF.INFO")
    f = open(fi, "w")
    for s in surf:
        enp, contc = smart_converge(mat=s)
        strt = Structure.from_file(contc)
        m = strt.lattice.matrix
        if s.comment.split("@")[0] == "sbulk":

            gs_energy = enp
            print("in surf_energy gs_energy for", s.comment, gs_energy)
        else:
            surf_en = (
                16
                * (
                    float(enp) * (strt.composition.num_atoms)
                    - float(strt.composition.num_atoms) * float(gs_energy)
                )
                / (2 * np.linalg.norm(np.cross(m[0], m[1])))
            )
            print("in surf_energy surf_en for", s.comment, surf_en)
            surf_list.append(surf_en)
            print(s.comment, "=", surf_en)
            line = str(s.comment) + str("=") + str(surf_en) + "\n"
            f.write(line)
    f.close()
    return surf_list, surf_header_list


def make_big(poscar=None, size=11.0):
    """
    Helper function to make supercell
 
    Args:
      poscar: Poscar object
      size: simulation size in Angstrom
    Returns:
       big: Poscar supercell object
    """
    struct = poscar.structure
    comm = poscar.comment
    a, b, c = struct.lattice.abc
    struct.make_supercell(
        [
            int(float(size) / float(a)) + 1,
            int(float(size) / float(b)) + 1,
            int(float(size) / float(c)) + 1,
        ]
    )
    big = Poscar(struct)
    big.comment = str(comm)
    return big


def main_func(mpid="", jid="", mat=None, enforc_cvn=False):
    """
    Main function to carry out property calculations

    Args:
        mpid: materialsproject id
        jid: jarvis-dft id
        mat: Poscar object
        enforc_cvn:  whether or not enforce conventional cell input structure
    Returns:
       en: final energy
       final: final structure
    """

    if mpid != "" or jid != "":
        data = loadfn(json_dat, cls=MontyDecoder)
        for d in data:
            mmpid = str(d["mpid"])
            jjid = str(d["mpid"])
            if mmpid == mpid:
                fin = d["structure"]
                break
            if jjpid == jid:
                fin = d["structure"]
                break

        strt = fin
        sg_mat = SpacegroupAnalyzer(strt)
        mat_cvn = sg_mat.get_conventional_standard_structure()
        mat_cvn.sort()
        if (
            int(strt.composition._natoms) == int(mat_cvn.composition._natoms)
            and enforc_cvn == True
        ):
            mat = Poscar(mat_cvn)
        else:
            mat = Poscar(strt)

        mpid = mpid.replace("-", "_")
        mpid = str("bulk@") + str(mpid)
        mat.comment = mpid

    en, final = smart_converge(mat=mat)
    print(en, final)


# if __name__ == '__main__':
# struct=Structure.from_file("POSCAR")
# pos=Poscar.from_file('POSCAR')
# #pos.comment='Surf-mp-C'
# main_func(mat=pos)



"""
Run and analyze BoltzTrap
"""
from __future__ import unicode_literals, print_function
from shutil import which
from pymatgen.electronic_structure.boltztrap import BoltztrapAnalyzer, BoltztrapRunner
from pymatgen.electronic_structure.plotter import BSPlotter, BoltztrapPlotter
from pymatgen.io.vasp.outputs import Vasprun
import matplotlib.pyplot as plt

plt.switch_backend("agg")
import json, os
from monty.serialization import MontyEncoder, loadfn, MontyDecoder


def boltz_run(vrun=""):
    """
  Helper function to run and store Bolztrap 
  Please note high denske k-point mesh is needed
  The automatic k-point convergene in JARVIS-DFT is generally good enough
  
  Args:
      vrun: path to vasprun.xml
  Returns:
        BoltztrapAnalyzer object
  """
    if which("x_trans") is None or which("BoltzTraP") is None:
        print("Please install BoltzTrap")
    v = Vasprun(vrun)
    kp = vrun.replace("vasprun.xml", "KPOINTS")
    out = vrun.replace("vasprun.xml", "OUTCAR")
    for line in open(out, "r"):
        if "NELECT" in line:
            nelect = float(line.split()[2])
    bs = v.get_band_structure(kp, line_mode=False)
    brun = BoltztrapRunner(bs, nelect)
    path = vrun.split("vasprun.xml")[0]
    folder = str(path) + str("/boltztrap")
    out_trans = str(folder) + str("/boltztrap.outputtrans")
    if not os.path.exists(out_trans):
        brun.run(path_dir=path)
        print("doesnt exist", out_trans)
    bana = BoltztrapAnalyzer.from_files(folder)
    return bana


def get_prop(bzana, prop="seebeck", dop_type="n", temp=700, dop=1e20, output="eigs"):
    """
    Helper function to obtain thermoelectric properties at specific values of temp,doping
    Args:
       bzana: BoltztrapAnalyzer object
       prop: one of 'seebeck','power_factor', 'electronic_conductivity', 'el_thermal_conductivity', 'zt'
       dop_type: 'n' or 'p' 
       temp: temperature in K
       dop: doping conc. in /cm3
       output: 'eigs' or 'tensor'
    Returns:
          val: desired values
    """
    doping = bzana.doping["p"]
    index = None
    for i, ii in enumerate(doping):
        if ii == dop:
            index = i
    if index is not None:
        if prop == "seebeck":
            val = bzana.get_seebeck(output=output)[dop_type][temp][index]
            return val
        if prop == "power_factor":
            val = bzana.get_power_factor(output=output)[dop_type][temp][index]
            return val
        if prop == "electronic_conductivity":
            val = bzana.get_conductivity(output=output)[dop_type][temp][index]
            return val
        if prop == "el_thermal_conductivity":
            val = bzana.get_thermal_conductivity(output=output)[dop_type][temp][index]
            return val
        if prop == "zt":
            val = bzana.get_zt(output=output)[dop_type][temp][index]
            return val
    else:
        print(
            "Value not available in the analyzer, run boltztrap with appropriate settings"
        )


if __name__ == "__main__":
    vrun = path = str(
        os.path.join(
            os.path.dirname(__file__),
            "../vasp/examples/SiOptb88/MAIN-RELAX-bulk@mp_149/vasprun.xml",
        )
    )
    b = boltz_run(vrun)
    val = get_prop(b, prop="zt")
    print(val)
